import unittest
from pathlib import Path
from typing import Any

import yaml
from pydantic import BaseModel, ConfigDict, Field

from jupyter_deploy.variables_config import (
    VARIABLES_CONFIG_V1_KEYS_ORDER,
    JupyterDeployVariablesConfigV1,
)


# Create a test version of the class without the validators for testing
class TestVariablesConfigV1NoValidators(BaseModel):
    model_config = ConfigDict(extra="allow")
    schema_version: int = 1  # Using Literal[1] would make sense here but int is fine for tests
    required: dict[str, Any] = Field(default_factory=dict)
    required_sensitive: dict[str, Any] = Field(default_factory=dict)
    overrides: dict[str, Any] = Field(default_factory=dict)
    defaults: dict[str, Any] = Field(default_factory=dict)


class TestJupyterDeployVariablesConfigV1(unittest.TestCase):
    variables_v1_content: str
    variables_v1_parsed_content: Any
    variables_initial_v1_content: str
    variables_initial_v1_parsed_content: Any

    @classmethod
    def setUpClass(cls) -> None:
        mock_variables_path = Path(__file__).parent / "mock_variables.yaml"
        with open(mock_variables_path) as f:
            cls.variables_v1_content = f.read()
        cls.variables_v1_parsed_content = yaml.safe_load(cls.variables_v1_content)

        mock_variables_initial_path = Path(__file__).parent / "mock_variables_initial.yaml"
        with open(mock_variables_initial_path) as f:
            cls.variables_initial_v1_content = f.read()
        cls.variables_initial_v1_parsed_content = yaml.safe_load(cls.variables_initial_v1_content)

    def test_can_parse_variables_v1(self) -> None:
        JupyterDeployVariablesConfigV1(
            **self.variables_v1_parsed_content  # type: ignore
        )

    def test_variables_v1_schema_version(self) -> None:
        variables = TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )
        self.assertEqual(variables.schema_version, 1)

    def test_can_parse_variables_initial_v1(self) -> None:
        # here we ensure that we can parse when override has no key (hence set to None)
        JupyterDeployVariablesConfigV1(
            **self.variables_initial_v1_parsed_content  # type: ignore
        )

    def test_variables_v1_required(self) -> None:
        variables = TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )
        self.assertEqual(variables.required["region"], "us-west-2")
        self.assertIsNone(variables.required["bucket_name"])
        self.assertIsNone(variables.required["storage_region"])

    def test_variables_v1_required_sensitive(self) -> None:
        variables = TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )
        self.assertEqual(variables.required_sensitive["aws_access_key"], "dummy-access-key")
        self.assertIsNone(variables.required_sensitive["aws_secret_key"])
        self.assertIsNone(variables.required_sensitive["api_token"])

    def test_variables_v1_overrides(self) -> None:
        variables = TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )
        self.assertEqual(variables.overrides["deployment_type"], "t3.large")
        self.assertEqual(variables.overrides["storage_size"], 100)

    def test_variables_v1_defaults(self) -> None:
        variables = TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )
        self.assertEqual(variables.defaults["deployment_type"], "t3.medium")
        self.assertEqual(variables.defaults["storage_size"], 50)
        self.assertEqual(variables.defaults["server_name"], "jupyter-server")
        self.assertEqual(variables.defaults["enable_monitoring"], True)

    def test_check_overrides_exist_pass(self) -> None:
        # Should not raise an error with valid overrides
        # We use a test class with overridden validators to allow us to test the basic parsing
        TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )

    def test_check_overrides_exist_fail(self) -> None:
        # Modify content to include unknown override
        invalid_content = self.variables_v1_parsed_content.copy()
        invalid_content["overrides"] = {"non_existent_key": "value", **invalid_content["overrides"]}

        with self.assertRaises(ValueError) as context:
            JupyterDeployVariablesConfigV1(**invalid_content)  # type: ignore

        self.assertIn("Unrecognized overrides", str(context.exception))
        self.assertIn("non_existent_key", str(context.exception))

    def test_check_no_var_name_repeat_pass(self) -> None:
        # Should not raise an error with no repeated variables
        # We use a test class with overridden validators to allow us to test the basic parsing
        TestVariablesConfigV1NoValidators(
            **self.variables_v1_parsed_content  # type: ignore
        )

    def test_check_no_var_name_repeat_fail_required_sensitive(self) -> None:
        # Modify content to include variable in both required and required_sensitive
        invalid_content = self.variables_v1_parsed_content.copy()
        invalid_content["required_sensitive"]["region"] = "overlapping-key"

        with self.assertRaises(ValueError) as context:
            JupyterDeployVariablesConfigV1(**invalid_content)  # type: ignore

        self.assertIn("Variables definition conflict", str(context.exception))
        self.assertIn("region", str(context.exception))

    def test_check_no_var_name_repeat_fail_required_defaults(self) -> None:
        # Modify content to include variable in both required and defaults
        invalid_content = self.variables_v1_parsed_content.copy()
        invalid_content["required"]["server_name"] = "overlapping-key"

        with self.assertRaises(ValueError) as context:
            JupyterDeployVariablesConfigV1(**invalid_content)  # type: ignore

        self.assertIn("Variables definition conflict", str(context.exception))
        self.assertIn("server_name", str(context.exception))

    def test_check_no_var_name_repeat_fail_sensitive_defaults(self) -> None:
        # Modify content to include variable in both required_sensitive and defaults
        invalid_content = self.variables_v1_parsed_content.copy()
        invalid_content["required_sensitive"]["server_name"] = "overlapping-key"

        with self.assertRaises(ValueError) as context:
            JupyterDeployVariablesConfigV1(**invalid_content)  # type: ignore

        self.assertIn("Variables definition conflict", str(context.exception))
        self.assertIn("server_name", str(context.exception))

    def test_keys_order(self) -> None:
        # Test that VARIABLES_CONFIG_V1_KEYS_ORDER matches expected order
        expected_order = [
            "schema_version",
            "required",
            "required_sensitive",
            "overrides",
            "defaults",
        ]
        self.assertEqual(VARIABLES_CONFIG_V1_KEYS_ORDER, expected_order)



from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_outputs, tf_variables
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler
from jupyter_deploy.provider import manifest_command_runner as cmd_runner
from jupyter_deploy.provider.resolved_clidefs import StrResolvedCliParameter


class TeamsHandler(BaseProjectHandler):
    """Handler class to manage team access to a jupyter app."""

    _output_handler: EngineOutputsHandler
    _variable_handler: EngineVariablesHandler

    def __init__(self) -> None:
        """Instantiate the TeamsHandler."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._output_handler = tf_outputs.TerraformOutputsHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
            self._variable_handler = tf_variables.TerraformVariablesHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
        else:
            raise NotImplementedError(f"OutputsHandler implementation not found for engine: {self.engine}")

    def add_teams(self, teams: list[str]) -> None:
        """Allowlist the teams to access the Jupyter app."""
        command = self.project_manifest.get_command("teams.add")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )

        joined_teams = ",".join(teams)
        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "teams": StrResolvedCliParameter(parameter_name="teams", value=joined_teams),
                "action": StrResolvedCliParameter(parameter_name="action", value="add"),
                "category": StrResolvedCliParameter(parameter_name="category", value="teams"),
            },
        )
        if success:
            runner.update_variables(command)

    def remove_teams(self, teams: list[str]) -> None:
        """Remove the teams from the allowlist of the Jupyter app."""
        command = self.project_manifest.get_command("teams.remove")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )

        joined_teams = ",".join(teams)
        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "teams": StrResolvedCliParameter(parameter_name="teams", value=joined_teams),
                "action": StrResolvedCliParameter(parameter_name="action", value="remove"),
                "category": StrResolvedCliParameter(parameter_name="category", value="teams"),
            },
        )
        if success:
            runner.update_variables(command)

    def set_teams(self, teams: list[str]) -> None:
        """Replace the list of teams allowlisted to the Jupyter app."""
        command = self.project_manifest.get_command("teams.set")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        joined_teams = ",".join(teams)

        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "teams": StrResolvedCliParameter(parameter_name="teams", value=joined_teams),
                "action": StrResolvedCliParameter(parameter_name="action", value="set"),
                "category": StrResolvedCliParameter(parameter_name="category", value="teams"),
            },
        )
        if success:
            runner.update_variables(command)

    def list_teams(self) -> list[str]:
        """Return a list of teams allowlisted to access the Jupyter app."""
        command = self.project_manifest.get_command("teams.list")
        console = self.get_console()

        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )

        success, _ = runner.run_command_sequence(
            command,
            cli_paramdefs={
                "category": StrResolvedCliParameter(parameter_name="category", value="teams"),
            },
        )
        if success:
            # No variables to update for list command
            return runner.get_result_value(command, "teams.list", list[str])
        return []



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from pydantic import ValidationError
from rich import console as rich_console

from jupyter_deploy import constants, fs_utils
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.handlers import base_project_handler
from jupyter_deploy.manifest import JupyterDeployManifest
from jupyter_deploy.variables_config import (
    VARIABLES_CONFIG_V1_COMMENTS,
    VARIABLES_CONFIG_V1_KEYS_ORDER,
    JupyterDeployVariablesConfig,
    JupyterDeployVariablesConfigV1,
)


class EngineVariablesHandler(ABC):
    def __init__(self, project_path: Path, project_manifest: JupyterDeployManifest) -> None:
        """Instantiate the base handler for the decorator."""
        self.project_path = project_path
        self.project_manifest = project_manifest
        self._variables_config: JupyterDeployVariablesConfig | None = None
        self._console: rich_console.Console | None = None

    def get_console(self) -> rich_console.Console:
        """Return the instance's rich console."""
        if self._console:
            return self._console
        self._console = rich_console.Console()
        return self._console

    def get_variables_config_path(self) -> Path:
        return self.project_path / constants.VARIABLES_FILENAME

    def _get_reset_variables_config(self) -> JupyterDeployVariablesConfig:
        """Retrieve the template variables, return reset variables config."""
        vardefs = self.get_template_variables()

        required: dict[str, Any] = {k: None for k, v in vardefs.items() if not v.has_default and not v.sensitive}
        sensitive: dict[str, Any] = {k: None for k, v in vardefs.items() if v.sensitive}
        defaults: dict[str, Any] = {k: v.default for k, v in vardefs.items() if v.has_default}
        return JupyterDeployVariablesConfigV1(
            schema_version=1,
            required=required,
            required_sensitive=sensitive,
            overrides={},
            defaults=defaults,
        )

    @property
    def variables_config(self) -> JupyterDeployVariablesConfig:
        if self._variables_config:
            return self._variables_config

        variables_config_path = self.project_path / constants.VARIABLES_FILENAME
        try:
            variables_config = base_project_handler.retrieve_variables_config(variables_config_path)
            self._variables_config = variables_config
            return variables_config
        except FileNotFoundError:
            # the user has deleted their variables.yaml, reset it to a fallback
            console = self.get_console()
            console.rule("Invalid variables.yaml", style="red")
            console.print(
                f":warning: variables config not found at: {variables_config_path.absolute()}", style="yellow"
            )
            console.line()
            console.rule(style="red")
            reset_variables_config = self._get_reset_variables_config()
            self._variables_config = reset_variables_config
            return self._variables_config
        except base_project_handler.NotADictError:
            # the user has corrupted their variables.yaml, reset to a fallback
            console = self.get_console()
            console.rule("Invalid variables.yaml", style="red")
            console.print(":warning: variables config was not a dict, resetting...", style="yellow")
            console.line()
            console.rule(style="red")
            reset_variables_config = self._get_reset_variables_config()
            self._variables_config = reset_variables_config
            return self._variables_config
        except ValidationError as e:
            # the user has corrupted their variables.yaml, reset to a fallback;
            console = self.get_console()
            console.rule("Invalid variables.yaml")
            console.print(":warning: variables config is invalid, resetting...", style="yellow")
            console.line()
            console.print(e, style="red")
            console.rule(style="red")
            reset_variables_config = self._get_reset_variables_config()
            self._variables_config = reset_variables_config
            return self._variables_config

    @abstractmethod
    def is_template_directory(self) -> bool:
        """Return True if the directory corresponds to a jupyter-deploy directory."""
        pass

    @abstractmethod
    def get_template_variables(self) -> dict[str, TemplateVariableDefinition]:
        """Return the dict of variable-name->variable-definition.

        This operation presumably requires file system operations and should
        be cached within each VariableHandler.
        """
        pass

    @abstractmethod
    def update_variable_records(self, varvalues: dict[str, Any], sensitive: bool = False) -> None:
        """Update the recorded values of all variables passed.

        Raises:
            KeyError if any of the variable name is not found
            TypeError if the any of the variable definition is not of the right type.
        """
        pass

    def sync_engine_varfiles_with_project_variables_config(self) -> None:
        """Update engine specific variable files from the variables config.

        Bypass all variables set to `null`.
        """

        required = self.variables_config.required
        sensitive = self.variables_config.required_sensitive
        overrides = self.variables_config.overrides
        varvalues: dict[str, Any] = {}
        sensitive_varvalues: dict[str, Any] = {}

        for var_name, var_value in required.items():
            if var_value is None:
                continue
            varvalues[var_name] = var_value

        for var_name, var_value in overrides.items():
            if var_value is None:
                continue
            varvalues[var_name] = var_value

        for sensitive_var_name, sensitive_var_value in sensitive.items():
            if sensitive_var_value is None:
                continue
            sensitive_varvalues[sensitive_var_name] = sensitive_var_value

        self.update_variable_records(varvalues)
        self.update_variable_records(sensitive_varvalues, sensitive=True)

    def get_variable_names_assigned_in_config(self) -> list[str]:
        """Return the variable names for which the user specified a value in `variables.yaml`."""
        assigned_variable_names: list[str] = []
        assigned_variable_names.extend([k for k, v in self.variables_config.required.items() if v is not None])
        assigned_variable_names.extend(
            [k for k, v in self.variables_config.required_sensitive.items() if v is not None]
        )
        assigned_variable_names.extend([k for k, v in self.variables_config.overrides.items() if v is not None])

        return assigned_variable_names

    def sync_project_variables_config(self, updated_values: dict[str, Any]) -> None:
        """Update the project variables.yaml to match the values."""

        curr_vars = self.variables_config
        new_required_dict = curr_vars.required.copy()
        new_sensitive_dict = curr_vars.required_sensitive.copy()
        new_overrides_dict = curr_vars.overrides.copy()

        for var_name, var_value in updated_values.items():
            if var_name in new_required_dict:
                new_required_dict[var_name] = var_value
            elif var_name in new_sensitive_dict:
                new_sensitive_dict[var_name] = var_value
            elif var_value is not None:  # only pass non-None values for overrides
                new_overrides_dict[var_name] = var_value

        new_variables_config = JupyterDeployVariablesConfigV1(
            schema_version=1,
            required=new_required_dict,
            required_sensitive=new_sensitive_dict,
            overrides=new_overrides_dict,
            defaults=curr_vars.defaults,
        )

        variables_config_path = self.project_path / constants.VARIABLES_FILENAME
        fs_utils.write_yaml_file_with_comments(
            variables_config_path,
            new_variables_config.model_dump(),
            key_order=VARIABLES_CONFIG_V1_KEYS_ORDER,
            comments=VARIABLES_CONFIG_V1_COMMENTS,
        )
        self._variables_config = new_variables_config

    def reset_recorded_variables(self) -> None:
        """Reset non-sensitive variables to their original values."""
        new_variables_config = JupyterDeployVariablesConfigV1(
            schema_version=1,
            required={k: None for k in self.variables_config.required},
            required_sensitive={k: None for k in self.variables_config.required_sensitive},
            overrides={},
            defaults=self.variables_config.defaults,
        )

        variables_config_path = self.project_path / constants.VARIABLES_FILENAME
        fs_utils.write_yaml_file_with_comments(
            variables_config_path,
            new_variables_config.model_dump(),
            key_order=VARIABLES_CONFIG_V1_KEYS_ORDER,
            comments=VARIABLES_CONFIG_V1_COMMENTS,
        )
        self._variables_config = new_variables_config

    def reset_recorded_secrets(self) -> None:
        """Reset sensitive variables to their original values."""
        variables_config = self.variables_config
        new_variables_config = JupyterDeployVariablesConfigV1(
            schema_version=1,
            required=variables_config.required,
            required_sensitive={k: None for k in variables_config.required_sensitive},
            overrides=variables_config.overrides,
            defaults=variables_config.defaults,
        )

        variables_config_path = self.project_path / constants.VARIABLES_FILENAME
        fs_utils.write_yaml_file_with_comments(
            variables_config_path,
            new_variables_config.model_dump(),
            key_order=VARIABLES_CONFIG_V1_KEYS_ORDER,
            comments=VARIABLES_CONFIG_V1_COMMENTS,
        )
        self._variables_config = new_variables_config



import unittest
from unittest.mock import Mock, patch

from rich.console import Console

from jupyter_deploy.provider.aws.aws_runner import AwsApiRunner, AwsService
from jupyter_deploy.provider.resolved_argdefs import ResolvedInstructionArgument
from jupyter_deploy.provider.resolved_resultdefs import ResolvedInstructionResult


class TestAwsApiRunner(unittest.TestCase):
    def test_init_no_region(self) -> None:
        runner = AwsApiRunner(region_name=None)
        self.assertIsNone(runner.region_name)
        self.assertEqual(runner.service_runners, {})

    def test_init_with_region(self) -> None:
        runner = AwsApiRunner(region_name="us-west-2")
        self.assertEqual(runner.region_name, "us-west-2")
        self.assertEqual(runner.service_runners, {})

    @patch("jupyter_deploy.provider.aws.aws_runner.AwsSsmRunner")
    def test_execute_instantiate_ssm_runner_and_call_execute(self, mock_ssm_runner_class: Mock) -> None:
        # Setup
        mock_ssm_runner = Mock()
        mock_ssm_runner_class.return_value = mock_ssm_runner

        expected_result = {"result_key": Mock(spec=ResolvedInstructionResult)}
        mock_ssm_runner.execute_instruction.return_value = expected_result

        runner = AwsApiRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {"arg1": Mock(spec=ResolvedInstructionArgument)}

        # Execute
        result = runner.execute_instruction(
            instruction_name="aws.ssm.command", resolved_arguments=resolved_args, console=console
        )

        # Assert
        mock_ssm_runner_class.assert_called_once_with(region_name="us-west-2")
        mock_ssm_runner.execute_instruction.assert_called_once_with(
            instruction_name="command", resolved_arguments=resolved_args, console=console
        )
        self.assertEqual(result, expected_result)

    @patch("jupyter_deploy.provider.aws.aws_runner.AwsEc2Runner")
    def test_execute_instantiate_ec2_runner_and_call_execute(self, mock_ec2_runner_class: Mock) -> None:
        # Setup
        mock_ec2_runner = Mock()
        mock_ec2_runner_class.return_value = mock_ec2_runner

        expected_result = {"result_key": Mock(spec=ResolvedInstructionResult)}
        mock_ec2_runner.execute_instruction.return_value = expected_result

        runner = AwsApiRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {"arg1": Mock(spec=ResolvedInstructionArgument)}

        # Execute
        result = runner.execute_instruction(
            instruction_name="aws.ec2.start-instance", resolved_arguments=resolved_args, console=console
        )

        # Assert
        mock_ec2_runner_class.assert_called_once_with(region_name="us-west-2")
        mock_ec2_runner.execute_instruction.assert_called_once_with(
            instruction_name="start-instance", resolved_arguments=resolved_args, console=console
        )
        self.assertEqual(result, expected_result)

    @patch("jupyter_deploy.provider.aws.aws_runner.AwsSsmRunner")
    def test_execute_recycle_service_runner(self, mock_ssm_runner_class: Mock) -> None:
        # Setup
        mock_ssm_runner = Mock()
        mock_ssm_runner_class.return_value = mock_ssm_runner

        runner = AwsApiRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {"arg1": Mock(spec=ResolvedInstructionArgument)}

        # Execute first instruction
        runner.execute_instruction(
            instruction_name="aws.ssm.command1", resolved_arguments=resolved_args, console=console
        )

        # Execute second instruction
        runner.execute_instruction(
            instruction_name="aws.ssm.command2", resolved_arguments=resolved_args, console=console
        )

        # Assert
        mock_ssm_runner_class.assert_called_once_with(region_name="us-west-2")
        self.assertEqual(mock_ssm_runner.execute_instruction.call_count, 2)
        self.assertEqual(len(runner.service_runners), 1)
        self.assertIn(AwsService.SSM, runner.service_runners)
        self.assertEqual(runner.service_runners[AwsService.SSM], mock_ssm_runner)

    def test_execute_raise_not_implemented_error_on_unmatches_service(self) -> None:
        # Setup
        runner = AwsApiRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {"arg1": Mock(spec=ResolvedInstructionArgument)}

        # Execute and Assert
        with self.assertRaises(NotImplementedError) as context:
            runner.execute_instruction(
                instruction_name="aws.unknown-service.command", resolved_arguments=resolved_args, console=console
            )

        self.assertIn("unknown-service", str(context.exception))

    def test_execute_raise_value_error_on_invalid_instruction_name(self) -> None:
        # Setup
        runner = AwsApiRunner(region_name="us-west-2")
        console = Mock(spec=Console)
        resolved_args: dict[str, ResolvedInstructionArgument] = {"arg1": Mock(spec=ResolvedInstructionArgument)}

        # Test cases for invalid instruction names
        invalid_instructions = [
            "",
            "..",
            "aws",  # Missing service and instruction
            "aws.",  # Missing service and instruction
            "aws.ssm.",  # Missing instruction
            "aws.ssm....",  # Missing instruction
        ]

        for invalid_instruction in invalid_instructions:
            with self.subTest(invalid_instruction=invalid_instruction):
                # Execute and Assert
                with self.assertRaises(ValueError) as context:
                    runner.execute_instruction(
                        instruction_name=invalid_instruction, resolved_arguments=resolved_args, console=console
                    )

                self.assertIn(invalid_instruction, str(context.exception))



import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from jupyter_deploy.handlers.project.config_handler import ConfigHandler
from jupyter_deploy.manifest import JupyterDeployManifestV1


class TestConfigHandler(unittest.TestCase):
    def get_mock_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return mocked config handler."""
        mock_handler = Mock()
        mock_has_recorded_variables = Mock()
        mock_verify_preset = Mock()
        mock_list_presets = Mock()
        mock_verify = Mock()
        mock_reset_recorded_variables = Mock()
        mock_reset_recorded_secrets = Mock()
        mock_configure = Mock()
        mock_record = Mock()

        mock_handler.has_recorded_variables = mock_has_recorded_variables
        mock_handler.verify_preset_exists = mock_verify_preset
        mock_handler.list_presets = mock_list_presets
        mock_handler.verify_requirements = mock_verify
        mock_handler.reset_recorded_variables = mock_reset_recorded_variables
        mock_handler.reset_recorded_secrets = mock_reset_recorded_secrets
        mock_handler.configure = mock_configure
        mock_handler.record = mock_record

        mock_has_recorded_variables.return_value = False
        mock_verify.return_value = True
        mock_verify_preset.return_value = True
        mock_configure.return_value = True
        mock_list_presets.return_value = ["all", "base", "none"]

        return (
            mock_handler,
            {
                "has_recorded_variables": mock_has_recorded_variables,
                "verify_requirements": mock_verify,
                "verify_preset_exists": mock_verify_preset,
                "list_presets": mock_list_presets,
                "reset_recorded_variables": mock_reset_recorded_variables,
                "reset_recorded_secrets": mock_reset_recorded_secrets,
                "configure": mock_configure,
                "record": mock_record,
            },
        )

    def setUp(self) -> None:
        self.mock_manifest = JupyterDeployManifestV1(
            **{  # type: ignore
                "schema_version": 1,
                "template": {
                    "name": "mock-template-name",
                    "engine": "terraform",
                    "version": "1.0.0",
                },
            }
        )

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    def test_config_handler_reads_the_manifest(self, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        handler = ConfigHandler()
        mock_retrieve_manifest.assert_called_once()
        self.assertEqual(handler.project_manifest, self.mock_manifest)
        self.assertEqual(handler.engine, self.mock_manifest.get_engine())

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    @patch("pathlib.Path.cwd")
    def test_config_handler_correctly_implements_tf_engine(
        self, mock_cwd: Mock, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        path = Path("/some/cur/dir")
        mock_cwd.return_value = path
        mock_retrieve_manifest.return_value = self.mock_manifest

        # right now, it defaults to terraform
        # in the future, it should infer it from the project
        handler = ConfigHandler()

        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_verify = tf_fns["verify_requirements"]
        tf_mock_configure = tf_fns["configure"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        self.assertIsNone(handler.preset_name)
        mock_tf_handler.assert_called_once_with(
            project_path=path, project_manifest=self.mock_manifest, output_filename=None
        )
        tf_mock_verify.assert_not_called()
        tf_mock_configure.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_validate_method_calls_underlying_verify_preset_method(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_has_recorded = tf_fns["has_recorded_variables"]
        tf_mock_verify_preset = tf_fns["verify_preset_exists"]
        tf_mock_list_presets = tf_fns["list_presets"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.validate_and_set_preset(preset_name="all")

        self.assertTrue(result)
        self.assertEqual(handler.preset_name, "all")
        tf_mock_has_recorded.assert_called_once()
        tf_mock_verify_preset.assert_called_once_with("all")
        tf_mock_list_presets.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_validate_method_does_not_verify_preset_when_passed_none(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_has_recorded = tf_fns["has_recorded_variables"]
        tf_mock_verify_preset = tf_fns["verify_preset_exists"]
        tf_mock_list_presets = tf_fns["list_presets"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.validate_and_set_preset(preset_name=None)

        self.assertTrue(result)
        self.assertIsNone(handler.preset_name)
        tf_mock_has_recorded.assert_called_once()
        tf_mock_verify_preset.assert_not_called()
        tf_mock_list_presets.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_validate_method_calls_list_presets_and_print_when_preset_is_not_found(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_verify_preset = tf_fns["verify_preset_exists"]
        tf_mock_list_presets = tf_fns["list_presets"]
        tf_mock_verify_preset.return_value = False
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.validate_and_set_preset(preset_name="i-do-not-exist")

        self.assertFalse(result)
        tf_mock_verify_preset.assert_called_once_with("i-do-not-exist")
        tf_mock_list_presets.assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_validate_method_calls_ignores_preset_when_detects_recorded_values(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_has_recorded = tf_fns["has_recorded_variables"]
        tf_mock_verify_preset = tf_fns["verify_preset_exists"]
        tf_mock_list_presets = tf_fns["list_presets"]
        tf_mock_has_recorded.return_value = True
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.validate_and_set_preset(preset_name="all")

        self.assertTrue(result)
        self.assertIsNone(handler.preset_name)
        tf_mock_has_recorded.assert_called_once()
        tf_mock_verify_preset.assert_not_called()
        tf_mock_list_presets.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_validate_method_calls_ignores_recorded_values_with_reset(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_has_recorded = tf_fns["has_recorded_variables"]
        tf_mock_verify_preset = tf_fns["verify_preset_exists"]
        tf_mock_list_presets = tf_fns["list_presets"]
        tf_mock_has_recorded.return_value = True
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.validate_and_set_preset(preset_name="all", will_reset_variables=True)

        self.assertTrue(result)
        self.assertEqual(handler.preset_name, "all")
        tf_mock_has_recorded.assert_not_called()
        tf_mock_verify_preset.assert_called_once()
        tf_mock_list_presets.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_verify_calls_underlying_handler_method(self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_verify = tf_fns["verify_requirements"]
        tf_mock_configure = tf_fns["configure"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.verify_requirements()

        self.assertTrue(result)
        tf_mock_verify.assert_called_once()
        tf_mock_configure.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_verify_surfaces_underlying_method_exception(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_verify = tf_fns["verify_requirements"]
        mock_tf_handler.return_value = tf_mock_handler_instance
        tf_mock_verify.side_effect = RuntimeError("some-error")

        handler = ConfigHandler()
        with self.assertRaisesRegex(RuntimeError, "some-error"):
            handler.verify_requirements()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_reset_variables_calls_underlying_handler_method(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        ConfigHandler()

        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_reset_vars = tf_fns["reset_recorded_variables"]
        tf_mock_reset_secrets = tf_fns["reset_recorded_secrets"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        handler.reset_recorded_variables()

        tf_mock_reset_vars.assert_called_once()
        tf_mock_reset_secrets.assert_not_called()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_reset_secrets_calls_underlying_handler_method(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_reset_vars = tf_fns["reset_recorded_variables"]
        tf_mock_reset_secrets = tf_fns["reset_recorded_secrets"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        handler.reset_recorded_secrets()

        tf_mock_reset_vars.assert_not_called()
        tf_mock_reset_secrets.assert_called_once()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_configure_calls_underlying_handler_method(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_verify = tf_fns["verify_requirements"]
        tf_mock_configure = tf_fns["configure"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        result = handler.configure()

        self.assertTrue(result)
        tf_mock_verify.assert_not_called()
        tf_mock_configure.assert_called_once_with(preset_name=None, variable_overrides=None)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_configure_passes_the_preset(self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_configure = tf_fns["configure"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        handler.validate_and_set_preset(preset_name="all")
        result = handler.configure()

        self.assertTrue(result)
        tf_mock_configure.assert_called_once_with(preset_name="all", variable_overrides=None)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_configure_passes_the_variables(self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_configure = tf_fns["configure"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        handler.validate_and_set_preset(preset_name="all")

        overrides = {"var1": Mock()}
        result = handler.configure(variable_overrides=overrides)  # type: ignore

        self.assertTrue(result)
        tf_mock_configure.assert_called_once_with(preset_name="all", variable_overrides=overrides)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_configure_surfaces_underlying_method_exception(
        self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock
    ) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_configure = tf_fns["configure"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        error = RuntimeError("another-error")
        tf_mock_configure.side_effect = error

        handler = ConfigHandler()

        with self.assertRaisesRegex(RuntimeError, "another-error"):
            handler.configure()

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_record_passes_the_flags_correctly(self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_record = tf_fns["record"]
        mock_tf_handler.return_value = tf_mock_handler_instance

        handler = ConfigHandler()
        handler.record()
        tf_mock_record.assert_called_once_with(record_vars=False, record_secrets=False)

        tf_mock_record.reset_mock()
        handler.record(record_vars=True)
        tf_mock_record.assert_called_once_with(record_vars=True, record_secrets=False)

        tf_mock_record.reset_mock()
        handler.record(record_secrets=True)
        tf_mock_record.assert_called_once_with(record_vars=False, record_secrets=True)

        tf_mock_record.reset_mock()
        handler.record(record_vars=True, record_secrets=True)
        tf_mock_record.assert_called_once_with(record_vars=True, record_secrets=True)

    @patch("jupyter_deploy.handlers.base_project_handler.retrieve_project_manifest")
    @patch("jupyter_deploy.engine.terraform.tf_config.TerraformConfigHandler")
    def test_record_surfaces_underlying_exception(self, mock_tf_handler: Mock, mock_retrieve_manifest: Mock) -> None:
        mock_retrieve_manifest.return_value = self.mock_manifest
        tf_mock_handler_instance, tf_fns = self.get_mock_handler_and_fns()
        tf_mock_record = tf_fns["record"]
        mock_tf_handler.return_value = tf_mock_handler_instance
        tf_mock_record.side_effect = RuntimeError("Cannot record!")

        handler = ConfigHandler()
        with self.assertRaises(RuntimeError):
            handler.record(record_vars=True)



from typing import Generic, TypeVar

from pydantic import BaseModel, ConfigDict

S = TypeVar("S")


class ResolvedCliParameter(BaseModel, Generic[S]):
    """Wrapper class for a value passed to a CLI command by the user."""

    model_config = ConfigDict(extra="allow")
    parameter_name: str
    value: S


class StrResolvedCliParameter(ResolvedCliParameter[str]):
    """Wrapper class for string value passed to a CLI command by the user."""

    pass


class IntResolvedCliParameter(ResolvedCliParameter[int]):
    """Wrapper class for int value passed to a CLI command by the user."""

    pass


class ListStrResolvedCliParameter(ResolvedCliParameter[list[str]]):
    """Wrapper class for list of strings value passed to a CLI command by the user."""

    pass



from abc import ABC, abstractmethod
from pathlib import Path

from jupyter_deploy.engine.enum import EngineType


class EngineUpHandler(ABC):
    def __init__(self, project_path: Path, engine: EngineType, engine_dir_path: Path) -> None:
        """Instantiate the base handler for `jd up` command."""
        self.project_path = project_path
        self.engine_dir_path = engine_dir_path
        self.engine = engine

    @abstractmethod
    def get_default_config_filename(self) -> str:
        pass

    @abstractmethod
    def apply(self, config_file_path: Path, auto_approve: bool = False) -> None:
        pass



from typing import Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.enum import InstructionArgumentSource, ResultSource, TransformType, UpdateSource, ValueSource


class InvalidServiceError(ValueError):
    """Special error class to indicate an invalid service name."""

    pass


class JupyterDeployTemplateV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str
    engine: str
    version: str


class JupyterDeployValueV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str
    source: str
    source_key: str = Field(alias="source-key")

    def get_source_type(self) -> ValueSource:
        """Return the declaration source type."""
        return ValueSource.from_string(self.source)


class JupyterDeployRequirementV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    name: str
    version: str | None = None


class JupyterDeployInstructionArgumentV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    api_attribute: str = Field(alias="api-attribute")
    source: str
    source_key: str = Field(alias="source-key")

    def get_source_type(self) -> InstructionArgumentSource:
        """Return the instruction argument source type."""
        return InstructionArgumentSource.from_string(self.source)


class JupyterDeployInstructionResultV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    result_name: str = Field(alias="result-name")
    source: str
    source_key: str = Field(alias="source-key")
    transform: str | None = None

    def get_source_type(self) -> ResultSource:
        """Return the instruction argument source type."""
        return ResultSource.from_string(self.source)

    def get_transform_type(self) -> TransformType:
        """Return the transform type to apply to the source."""
        return TransformType.from_string(self.transform)


class JupyterDeployCommandUpdateV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    variable_name: str = Field(alias="variable-name")
    source: str
    source_key: str = Field(alias="source-key")
    transform: str | None = None

    def get_source_type(self) -> UpdateSource:
        """Return the instruction argument source type."""
        return UpdateSource.from_string(self.source)

    def get_transform_type(self) -> TransformType:
        """Return the transform type to apply to the source."""
        return TransformType.from_string(self.transform)


class JupyterDeployInstructionV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    api_name: str = Field(alias="api-name")
    arguments: list[JupyterDeployInstructionArgumentV1]


class JupyterDeployCommandV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    cmd: str
    sequence: list[JupyterDeployInstructionV1]
    results: list[JupyterDeployInstructionResultV1] | None = None
    updates: list[JupyterDeployCommandUpdateV1] | None = None


class JupyterDeployManifestV1(BaseModel):
    model_config = ConfigDict(extra="allow")
    schema_version: Literal[1]
    template: JupyterDeployTemplateV1
    requirements: list[JupyterDeployRequirementV1] | None = None
    values: list[JupyterDeployValueV1] | None = None
    services: list[str] | None = None
    commands: list[JupyterDeployCommandV1] | None = None

    def get_engine(self) -> EngineType:
        """Return the engine type."""
        return EngineType.from_string(self.template.engine)

    def get_declared_value(self, value_name: str) -> JupyterDeployValueV1:
        """Return the declared value definition.

        Raises:
            NotImplementedError if the manifest has no declared values.
            NotImplementedError if the value is not found.
        """
        value = next((val for val in (self.values or []) if val.name == value_name), None)
        if not value:
            raise NotImplementedError(f"No declaration found for value: {value_name}")
        return value

    def get_command(self, cmd_name: str) -> JupyterDeployCommandV1:
        """Return the command details.

        Raises:
            NotImplementedError if the manifest has no declared command.
            NotImplementedError if the command is not found.
        """
        command = next((cmd for cmd in (self.commands or []) if cmd.cmd == cmd_name), None)
        if not command:
            raise NotImplementedError(f"No implementation found for command: {cmd_name}")
        return command

    def get_services(self) -> list[str]:
        """Return the services name."""
        if not self.services:
            return []
        return self.services

    def get_validated_service(self, svc: str, allow_all: bool = True) -> str:
        """Return the value matching the service.

        Raises:
            InvalidServiceError if service is invalid
        """
        services = self.get_services()

        # no services defined: just allow
        if not services:
            return svc

        # first, if service is explicitely listed in services, return it
        if svc in services:
            return svc

        # else, use placeholders
        if svc == "default":
            if len(services):
                return services[0]
        elif svc == "all" and allow_all:
            return "all"

        raise InvalidServiceError(f"Invalid service, use one of {services}")

    def has_command(self, cmd_name: str) -> bool:
        """Return true if the manifest defines the command, false otherwise."""
        command = next((cmd for cmd in (self.commands or []) if cmd.cmd == cmd_name), None)
        return command is not None

    def get_requirements(self) -> list[JupyterDeployRequirementV1]:
        """Return the list of requirements as declared in the manifest."""
        return self.requirements or []


# Combined type using discriminated union
JupyterDeployManifest = Annotated[JupyterDeployManifestV1, "schema_version"]



import unittest
from unittest.mock import Mock, patch

from typer.testing import CliRunner

from jupyter_deploy.cli.teams_app import teams_app


class TestTeamsApp(unittest.TestCase):
    """Test cases for the teams_app module."""

    def test_help_command(self) -> None:
        """Test the help command."""
        self.assertTrue(len(teams_app.info.help or "") > 0, "help should not be empty")

        runner = CliRunner()
        result = runner.invoke(teams_app, ["--help"])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(result.stdout.index("add") > 0)
        self.assertTrue(result.stdout.index("remove") > 0)
        self.assertTrue(result.stdout.index("set") > 0)
        self.assertTrue(result.stdout.index("list") > 0)

    def test_no_arg_defaults_to_help(self) -> None:
        """Test that running the app with no arguments shows help."""
        runner = CliRunner()
        result = runner.invoke(teams_app, [])

        self.assertEqual(result.exit_code, 0)
        self.assertTrue(len(result.stdout) > 0)


class TestTeamAddCmd(unittest.TestCase):
    def get_mock_teams_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock teams handler."""
        mock_teams_handler = Mock()
        mock_add_teams = Mock()
        mock_get_console = Mock()

        mock_teams_handler.add_teams = mock_add_teams
        mock_teams_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_teams_handler, {
            "add_teams": mock_add_teams,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_teams_handler_and_calls_add_teams(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["add", "team1", "team2"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_teams_handler_class.assert_called_once()
        mock_handler_fns["add_teams"].assert_called_once_with(["team1", "team2"])

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_teams_handler_class: Mock) -> None:
        # Setup
        mock_teams_handler, _ = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["add", "team1", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_teams_handler_add_teams_raises(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_handler_fns["add_teams"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["add", "team1"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestTeamRemoveCmd(unittest.TestCase):
    def get_mock_teams_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock teams handler."""
        mock_teams_handler = Mock()
        mock_remove_teams = Mock()
        mock_get_console = Mock()

        mock_teams_handler.remove_teams = mock_remove_teams
        mock_teams_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_teams_handler, {
            "remove_teams": mock_remove_teams,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_teams_handler_and_calls_remove_teams(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that remove command instantiates TeamsHandler and calls remove_teams."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["remove", "team1", "team2"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_teams_handler_class.assert_called_once()
        mock_handler_fns["remove_teams"].assert_called_once_with(["team1", "team2"])

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_teams_handler_class: Mock) -> None:
        """Test that remove command switches directory when a project path is provided."""
        # Setup
        mock_teams_handler, _ = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["remove", "team1", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_teams_handler_remove_teams_raises(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that remove command propagates exceptions from remove_teams."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_handler_fns["remove_teams"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["remove", "team1"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestTeamSetCmd(unittest.TestCase):
    def get_mock_teams_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock teams handler."""
        mock_teams_handler = Mock()
        mock_set_teams = Mock()
        mock_get_console = Mock()

        mock_teams_handler.set_teams = mock_set_teams
        mock_teams_handler.get_console = mock_get_console

        mock_get_console.return_value = Mock()

        return mock_teams_handler, {
            "set_teams": mock_set_teams,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_teams_handler_and_calls_set_teams(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that set command instantiates TeamsHandler and calls set_teams."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["set", "team1", "team2"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_teams_handler_class.assert_called_once()
        mock_handler_fns["set_teams"].assert_called_once_with(["team1", "team2"])

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_teams_handler_class: Mock) -> None:
        """Test that set command switches directory when a project path is provided."""
        # Setup
        mock_teams_handler, _ = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["set", "team1", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_teams_handler_set_teams_raises(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that set command propagates exceptions from set_teams."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_handler_fns["set_teams"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["set", "team1"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)


class TestTeamListCmd(unittest.TestCase):
    def get_mock_teams_handler(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mock teams handler."""
        mock_teams_handler = Mock()
        mock_list_teams = Mock()
        mock_get_console = Mock()

        mock_teams_handler.list_teams = mock_list_teams
        mock_teams_handler.get_console = mock_get_console

        mock_list_teams.return_value = ["team1", "team2"]
        mock_get_console.return_value = Mock()

        return mock_teams_handler, {
            "list_teams": mock_list_teams,
            "get_console": mock_get_console,
        }

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_instantiates_teams_handler_and_calls_list_teams(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that list command instantiates TeamsHandler and calls list_teams."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["list"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_teams_handler_class.assert_called_once()
        mock_handler_fns["list_teams"].assert_called_once()

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_uses_handler_console_to_print_teams_list(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that list command uses the handler's console to print the list."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["list"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("team1, team2" in str(mock_call))

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_handles_no_teams(self, mock_project_dir: Mock, mock_teams_handler_class: Mock) -> None:
        """Test that list command handles the case when no teams are allowlisted."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_handler_fns["list_teams"].return_value = []

        mock_console = Mock()
        mock_handler_fns["get_console"].return_value = mock_console
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["list"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_console.print.assert_called_once()
        mock_call = mock_console.print.mock_calls[0]
        self.assertTrue("None" in str(mock_call))

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_switches_dir_when_passed_a_project(self, mock_project_dir: Mock, mock_teams_handler_class: Mock) -> None:
        """Test that list command switches directory when a project path is provided."""
        # Setup
        mock_teams_handler, _ = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["list", "--path", "/test/project/path"])

        # Assert
        self.assertEqual(result.exit_code, 0)
        mock_project_dir.assert_called_once_with("/test/project/path")

    @patch("jupyter_deploy.handlers.access.team_handler.TeamsHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_raises_when_teams_handler_list_teams_raises(
        self, mock_project_dir: Mock, mock_teams_handler_class: Mock
    ) -> None:
        """Test that list command propagates exceptions from list_teams."""
        # Setup
        mock_teams_handler, mock_handler_fns = self.get_mock_teams_handler()
        mock_teams_handler_class.return_value = mock_teams_handler
        mock_handler_fns["list_teams"].side_effect = Exception("Test error")
        mock_project_dir.return_value.__enter__.return_value = None

        # Execute
        runner = CliRunner()
        result = runner.invoke(teams_app, ["list"])

        # Assert
        self.assertNotEqual(result.exit_code, 0)



from pathlib import Path

from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_variables
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.manifest import JupyterDeployManifest


class VariablesHandler:
    """Base class to manage the variables of a jupyter-deploy project."""

    _handler: EngineVariablesHandler

    def __init__(self, project_path: Path, project_manifest: JupyterDeployManifest) -> None:
        """Instantiate the variables handler."""
        engine = project_manifest.get_engine()

        if engine == EngineType.TERRAFORM:
            self._handler = tf_variables.TerraformVariablesHandler(
                project_path=project_path, project_manifest=project_manifest
            )
        else:
            raise NotImplementedError(f"VariablesHandler implementation not found for engine: {engine}")

    def is_template_directory(self) -> bool:
        """Return True if the directory corresponds to a jupyter-deploy project."""
        return self._handler.is_template_directory()

    def get_template_variables(self) -> dict[str, TemplateVariableDefinition]:
        """Call underlying engine handler, return dict of var-name->var-definition."""
        if not self.is_template_directory():
            return {}
        return self._handler.get_template_variables()



import unittest
from pathlib import Path
from unittest.mock import MagicMock, patch

from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.template_utils import TEMPLATE_ENTRY_POINTS, get_templates


class TestTemplateUtils(unittest.TestCase):
    """Test class for template_utils module."""

    @patch("importlib.metadata.entry_points")
    def test_get_templates_success(self, mock_entry_points: MagicMock) -> None:
        """Test that get_templates correctly loads templates from entry points."""
        # Setup
        mock_entry_point1 = MagicMock()
        mock_entry_point1.name = "aws_ec2_base"
        mock_entry_point1.load.return_value = Path("/mock/template/path")

        mock_entry_point2 = MagicMock()
        mock_entry_point2.name = "aws_lambda_basic"
        mock_entry_point2.load.return_value = Path("/mock/lambda/path")

        mock_entry_points.return_value = [mock_entry_point1, mock_entry_point2]

        # Mock Path.exists to return True for our mock paths
        with patch("pathlib.Path.exists", return_value=True):
            # Execute
            templates = get_templates(EngineType.TERRAFORM)

            # Assert
            self.assertEqual(len(templates), 2)
            self.assertEqual(templates["aws:ec2:base"], Path("/mock/template/path"))
            self.assertEqual(templates["aws:lambda:basic"], Path("/mock/lambda/path"))
            mock_entry_points.assert_called_once_with(group=TEMPLATE_ENTRY_POINTS[EngineType.TERRAFORM])
            mock_entry_point1.load.assert_called_once()
            mock_entry_point2.load.assert_called_once()

    def test_get_templates_unsupported_engine(self) -> None:
        """Test that get_templates handles unsupported engine types."""
        # Execute
        # Using a string that's not an EngineType for testing error handling
        templates = get_templates("unsupported_engine")  # type: ignore

        # Assert
        self.assertEqual(len(templates), 0)

    @patch("importlib.metadata.entry_points")
    def test_get_templates_invalid_path(self, mock_entry_points: MagicMock) -> None:
        """Test that get_templates handles invalid paths."""
        # Setup
        mock_entry_point = MagicMock()
        mock_entry_point.name = "aws_ec2_base"
        mock_entry_point.load.return_value = "not_a_path"  # Invalid path

        mock_entry_points.return_value = [mock_entry_point]

        # Execute
        templates = get_templates(EngineType.TERRAFORM)

        # Assert
        self.assertEqual(len(templates), 0)
        mock_entry_points.assert_called_once_with(group=TEMPLATE_ENTRY_POINTS[EngineType.TERRAFORM])
        mock_entry_point.load.assert_called_once()

    @patch("importlib.metadata.entry_points")
    def test_get_templates_nonexistent_path(self, mock_entry_points: MagicMock) -> None:
        """Test that get_templates handles paths that don't exist."""
        # Setup
        mock_entry_point = MagicMock()
        mock_entry_point.name = "aws_ec2_base"
        mock_entry_point.load.return_value = Path("/nonexistent/path")

        mock_entry_points.return_value = [mock_entry_point]

        # Mock Path.exists to return False
        with patch("pathlib.Path.exists", return_value=False):
            # Execute
            templates = get_templates(EngineType.TERRAFORM)

            # Assert
            self.assertEqual(len(templates), 0)
            mock_entry_points.assert_called_once_with(group=TEMPLATE_ENTRY_POINTS[EngineType.TERRAFORM])
            mock_entry_point.load.assert_called_once()

    @patch("jupyter_deploy.template_utils.get_templates")
    def test_templates_loaded_for_all_engines(self, mock_get_templates: MagicMock) -> None:
        """Test that TEMPLATES dictionary is populated with all supported engines."""
        # Setup
        mock_get_templates.side_effect = lambda engine: {"test:template": Path(f"/mock/{engine}/path")}

        # Execute
        from jupyter_deploy.template_utils import TEMPLATE_ENTRY_POINTS

        templates = {engine: mock_get_templates(engine) for engine in TEMPLATE_ENTRY_POINTS}

        # Assert
        self.assertEqual(len(templates), len(TEMPLATE_ENTRY_POINTS))
        for engine in TEMPLATE_ENTRY_POINTS:
            self.assertIn(engine, templates)
            self.assertEqual(templates[engine], {"test:template": Path(f"/mock/{engine}/path")})

        # Verify get_templates was called once for each engine
        self.assertEqual(mock_get_templates.call_count, len(TEMPLATE_ENTRY_POINTS))



import unittest

from jupyter_deploy import type_utils


class TestTypeUtils(unittest.TestCase):
    """Tests for type_utils module."""

    def test_is_list_str_repr_with_valid_inputs(self) -> None:
        """Test is_list_str_repr() with valid inputs representing list of strings."""
        # Test with 'list' and 'string' - this is the most common case in practice
        self.assertTrue(type_utils.is_list_str_repr(["list", "string"]))

        # Test with 'list' and 'str' - another valid representation
        self.assertTrue(type_utils.is_list_str_repr(["list", "str"]))

        # Test with 'list' and 'dynamic' - empty array case
        self.assertTrue(type_utils.is_list_str_repr(["list", "dynamic"]))

        # Test with additional elements in the list
        self.assertTrue(type_utils.is_list_str_repr(["list", "string", "additional", "elements"]))
        self.assertTrue(type_utils.is_list_str_repr(["list", "str", True]))

    def test_is_list_str_repr_with_invalid_inputs(self) -> None:
        """Test is_list_str_repr() with invalid inputs that should return False."""
        # Test with non-list inputs
        self.assertFalse(type_utils.is_list_str_repr("list"))
        self.assertFalse(type_utils.is_list_str_repr(123))
        self.assertFalse(type_utils.is_list_str_repr(None))
        self.assertFalse(type_utils.is_list_str_repr({}))

        # Test with wrong first element
        self.assertFalse(type_utils.is_list_str_repr(["tuple", "string"]))
        self.assertFalse(type_utils.is_list_str_repr(["LIST", "string"]))  # Case sensitive check

        # Test with wrong second element
        self.assertFalse(type_utils.is_list_str_repr(["list", "int"]))
        self.assertFalse(type_utils.is_list_str_repr(["list", "number"]))
        self.assertFalse(type_utils.is_list_str_repr(["list", "bool"]))

        # Test with too short list
        self.assertFalse(type_utils.is_list_str_repr(["list"]))
        self.assertFalse(type_utils.is_list_str_repr([]))

        # Test with correct elements but in wrong order
        self.assertFalse(type_utils.is_list_str_repr(["string", "list"]))

        # Test with Python types instead of strings
        self.assertFalse(type_utils.is_list_str_repr(list))
        self.assertFalse(type_utils.is_list_str_repr(dict))
        self.assertFalse(type_utils.is_list_str_repr(set))
        self.assertFalse(type_utils.is_list_str_repr(int))

    def test_is_list_str_repr_with_real_terraform_output_format(self) -> None:
        """Test is_list_str_repr() with the exact format that Terraform would output."""
        # This is what we'd typically get from Terraform's output command
        terraform_output_type = ["list", "string"]
        self.assertTrue(type_utils.is_list_str_repr(terraform_output_type))

        # Sometimes Terraform might use 'str' instead
        terraform_output_type_alt = ["list", "str"]
        self.assertTrue(type_utils.is_list_str_repr(terraform_output_type_alt))

    def test_is_list_str_repr_with_edge_cases(self) -> None:
        """Test is_list_str_repr() with edge cases."""
        # Test with nested lists - these should fail
        self.assertFalse(type_utils.is_list_str_repr([["list"], "string"]))
        self.assertFalse(type_utils.is_list_str_repr(["list", ["string"]]))

        # Test with tuple instead of list
        self.assertFalse(type_utils.is_list_str_repr(("list", "string")))

        # Test with non-string elements but correct structure
        self.assertFalse(type_utils.is_list_str_repr([list, "string"]))
        self.assertFalse(type_utils.is_list_str_repr(["list", str]))



from jupyter_deploy.engine.engine_outputs import EngineOutputsHandler
from jupyter_deploy.engine.enum import EngineType
from jupyter_deploy.engine.terraform import tf_outputs, tf_variables
from jupyter_deploy.handlers.base_project_handler import BaseProjectHandler
from jupyter_deploy.provider import manifest_command_runner as cmd_runner


class HostHandler(BaseProjectHandler):
    """Handler class to directly interact with the host running jupyter server."""

    _output_handler: EngineOutputsHandler

    def __init__(self) -> None:
        """Instantiate the Host handler."""
        super().__init__()

        if self.engine == EngineType.TERRAFORM:
            self._output_handler = tf_outputs.TerraformOutputsHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
            self._variable_handler = tf_variables.TerraformVariablesHandler(
                project_path=self.project_path, project_manifest=self.project_manifest
            )
        else:
            raise NotImplementedError(f"OutputsHandler implementation not found for engine: {self.engine}")

    def get_host_status(self) -> str:
        """Returns the status of the host machine."""
        command = self.project_manifest.get_command("host.status")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(command, cli_paramdefs={})
        return runner.get_result_value(command, "host.status", str)

    def stop_host(self) -> None:
        """Stop the host machine."""
        command = self.project_manifest.get_command("host.stop")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(command, cli_paramdefs={})

    def start_host(self) -> None:
        """Start the host machine."""
        command = self.project_manifest.get_command("host.start")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(
            command,
            cli_paramdefs={},
        )

    def restart_host(self) -> None:
        """Restart the host machine."""
        command = self.project_manifest.get_command("host.restart")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(
            command,
            cli_paramdefs={},
        )

    def connect(self) -> None:
        """Start an SSH-style connection to the host."""
        command = self.project_manifest.get_command("host.connect")
        console = self.get_console()
        runner = cmd_runner.ManifestCommandRunner(
            console=console, output_handler=self._output_handler, variable_handler=self._variable_handler
        )
        runner.run_command_sequence(
            command,
            cli_paramdefs={},
        )



from pathlib import Path

from jupyter_deploy.engine.engine_open import EngineOpenHandler
from jupyter_deploy.engine.terraform import tf_outputs
from jupyter_deploy.manifest import JupyterDeployManifest


class TerraformOpenHandler(EngineOpenHandler):
    """Terraform implementation of the EngineOpenHandler."""

    def __init__(
        self,
        project_path: Path,
        project_manifest: JupyterDeployManifest,
    ) -> None:
        """Instantiate the terraform open handler."""
        outputs_handler = tf_outputs.TerraformOutputsHandler(
            project_path=project_path,
            project_manifest=project_manifest,
        )

        super().__init__(
            project_path=project_path,
            project_manifest=project_manifest,
            output_handler=outputs_handler,
        )



def is_list_str_repr(arg: object) -> bool:
    """Return True if the argument is a list[str]."""
    return isinstance(arg, list) and len(arg) >= 2 and arg[0] == "list" and arg[1] in ["string", "str", "dynamic"]



import json
from abc import ABC, abstractmethod
from typing import Generic, TypeVar  # noqa: UP035

from pydantic import BaseModel, ConfigDict

from jupyter_deploy.engine.terraform.tf_types import TerraformType
from jupyter_deploy.engine.vardefs import (
    AnyNumericTemplateVariableDefinition,
    BoolTemplateVariableDefinition,
    DictStrTemplateVariableDefinition,
    FloatTemplateVariableDefinition,
    IntTemplateVariableDefinition,
    ListMapStrTemplateVariableDefinition,
    ListStrTemplateVariableDefinition,
    StrTemplateVariableDefinition,
    TemplateVariableDefinition,
)

V = TypeVar("V")


class ParsedVariablesDotTf(BaseModel):
    model_config = ConfigDict(extra="allow")
    variable: list[dict[str, dict]]


class TerraformVariableDefinition(BaseModel, Generic[V], ABC):
    """Terraform-specific wrapper class for variable definition."""

    model_config = ConfigDict(extra="allow")
    variable_name: str
    tf_type: TerraformType
    description: str = ""
    default: V | None = None
    sensitive: bool = False
    has_default: bool = False

    @abstractmethod
    def to_template_definition(self) -> TemplateVariableDefinition:
        """Return the equivalent instance of the TemplateVariableDefinition subclass."""
        pass


class TerraformStrVariableDefinition(TerraformVariableDefinition[str]):
    """Terraform wrapper class for variable definition of type 'string'."""

    tf_type: TerraformType = TerraformType.STRING

    def to_template_definition(self) -> TemplateVariableDefinition:
        return StrTemplateVariableDefinition(**self.model_dump())


class TerraformNumberVariableDefinition(TerraformVariableDefinition[int | float]):
    """Terraform wrapper class for variable definition of type 'number'."""

    tf_type: TerraformType = TerraformType.NUMBER

    def to_template_definition(self) -> TemplateVariableDefinition:
        # terraform doesn't differentiate int from float in numeric value.
        # we attempt to infer it from default if defined.
        if isinstance(self.default, int):
            return IntTemplateVariableDefinition(**self.model_dump())
        elif isinstance(self.default, float):
            return FloatTemplateVariableDefinition(**self.model_dump())
        else:
            return AnyNumericTemplateVariableDefinition(**self.model_dump())


class TerraformBoolVariableDefinition(TerraformVariableDefinition[bool]):
    """Terraform wrapper class for variable definition of type 'bool'."""

    tf_type: TerraformType = TerraformType.BOOL

    def to_template_definition(self) -> TemplateVariableDefinition:
        return BoolTemplateVariableDefinition(**self.model_dump())


class TerraformListOfStrVariableDefinition(TerraformVariableDefinition[list[str]]):
    """Terraform wrapper class for variable definition of type 'list(string)'."""

    tf_type: TerraformType = TerraformType.LIST_STR

    def to_template_definition(self) -> TemplateVariableDefinition:
        return ListStrTemplateVariableDefinition(**self.model_dump())


class TerraformMapOfStrVariableDefinition(TerraformVariableDefinition[dict[str, str]]):
    """Terraform wrapper class for variable definition of type 'map(string)'."""

    tf_type: TerraformType = TerraformType.MAP_STR

    def to_template_definition(self) -> TemplateVariableDefinition:
        return DictStrTemplateVariableDefinition(**self.model_dump())


class TerraformListOfMapStrVariableDefinition(TerraformVariableDefinition[list[dict[str, str]]]):
    """Terraform wrapper class for variable definition of type 'list(map(string))'."""

    tf_type: TerraformType = TerraformType.LIST_MAP_STR

    def to_template_definition(self) -> TemplateVariableDefinition:
        return ListMapStrTemplateVariableDefinition(**self.model_dump())


def create_tf_variable_definition(parsed_config: dict) -> TerraformVariableDefinition:
    """Return an instance of the corresponding TerraformVariable class."""
    tf_type = parsed_config.get("tf_type")
    if tf_type == TerraformType.STRING:
        return TerraformStrVariableDefinition(**parsed_config)
    elif tf_type == TerraformType.NUMBER:
        return TerraformNumberVariableDefinition(**parsed_config)
    elif tf_type == TerraformType.BOOL:
        return TerraformBoolVariableDefinition(**parsed_config)
    elif tf_type == TerraformType.LIST_STR:
        return TerraformListOfStrVariableDefinition(**parsed_config)
    elif tf_type == TerraformType.MAP_STR:
        return TerraformMapOfStrVariableDefinition(**parsed_config)
    elif tf_type == TerraformType.LIST_MAP_STR:
        return TerraformListOfMapStrVariableDefinition(**parsed_config)
    raise NotImplementedError(f"No terraform class found for type: {tf_type}.")


def to_tf_var_option(var_def: TemplateVariableDefinition) -> list[str]:
    """Return the 'bar' value to pass to terraform as -var 'foo=bar'."""
    if var_def.assigned_value is None:
        return ["-var", f"{var_def.variable_name}=null"]

    if isinstance(var_def, StrTemplateVariableDefinition) and var_def.assigned_value == "":
        return ["-var", f'{var_def.variable_name}=""']

    if isinstance(var_def, BoolTemplateVariableDefinition):
        return ["-var", f"{var_def.variable_name}={'true' if var_def.assigned_value else 'false'}"]

    if isinstance(var_def, ListStrTemplateVariableDefinition):
        str_out = json.dumps(var_def.assigned_value)
        return ["-var", f"{var_def.variable_name}={str_out}"]

    if isinstance(var_def, DictStrTemplateVariableDefinition):
        str_out = json.dumps(var_def.assigned_value)
        return ["-var", f"{var_def.variable_name}={str_out}"]

    if isinstance(var_def, ListMapStrTemplateVariableDefinition):
        str_out = json.dumps(var_def.assigned_value)
        return ["-var", f"{var_def.variable_name}={str_out}"]

    return ["-var", f"{var_def.variable_name}={var_def.assigned_value}"]



import unittest
from pathlib import Path
from unittest.mock import Mock

from jupyter_deploy.engine.engine_down import EngineDownHandler
from jupyter_deploy.engine.outdefs import ListStrTemplateOutputDefinition


class EngineDownTester(EngineDownHandler):
    """Nested class for testing."""

    def destroy(self, auto_approve: bool = False) -> None:
        pass


class TestEngineDown(unittest.TestCase):
    def get_mock_outputs_handler_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mocked outputs handler."""
        mock_handler = Mock()
        mock_get_declared_output_def = Mock()
        mock_handler.get_declared_output_def = mock_get_declared_output_def

        mock_get_declared_output_def.return_value = ListStrTemplateOutputDefinition(
            output_name="persisting_resources", value=[]
        )

        return mock_handler, {"get_declared_output_def": mock_get_declared_output_def}

    def get_mock_console_and_fns(self) -> tuple[Mock, dict[str, Mock]]:
        """Return a mocked rich console instance."""
        mock_console = Mock()
        mock_print = Mock()
        mock_console.print = mock_print
        return mock_console, {"print": mock_print}

    def test_init(self) -> None:
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_manifest = Mock()
        handler = EngineDownTester(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )
        self.assertEqual(handler.project_path, Path("/project/path"))
        self.assertEqual(handler.project_manifest, mock_manifest)
        self.assertEqual(handler.output_handler, mock_outputs_handler)
        mock_outputs_handler_fns["get_declared_output_def"].assert_not_called()

    def test_get_persisting_resources_value_not_declared(self) -> None:
        # Setup
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].side_effect = NotImplementedError("Value not declared")
        mock_manifest = Mock()
        mock_manifest.get_engine.return_value = "terraform"
        handler = EngineDownTester(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Execute
        result = handler.get_persisting_resources()

        # Verify
        self.assertEqual(result, [])
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once_with(
            "persisting_resources", ListStrTemplateOutputDefinition
        )

    def test_get_persisting_happy_case(self) -> None:
        # Setup
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        expected_resources = [
            'aws_ebs_volume.additional_volumes["0"]',
            'aws_efs_file_system.additional_file_systems["0"]',
        ]
        mock_outputs_handler_fns["get_declared_output_def"].return_value = ListStrTemplateOutputDefinition(
            output_name="persisting_resources", value=expected_resources
        )
        mock_manifest = Mock()
        mock_manifest.get_engine.return_value = "terraform"
        handler = EngineDownTester(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Execute
        result = handler.get_persisting_resources()

        # Verify
        self.assertEqual(result, expected_resources)
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once_with(
            "persisting_resources", ListStrTemplateOutputDefinition
        )

    def test_get_persisting_resources_output_not_found(self) -> None:
        # Setup
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].side_effect = KeyError("Output not found")
        mock_manifest = Mock()
        mock_manifest.get_engine.return_value = "terraform"
        handler = EngineDownTester(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Execute
        result = handler.get_persisting_resources()

        # Verify
        self.assertEqual(result, [])
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once_with(
            "persisting_resources", ListStrTemplateOutputDefinition
        )

    def test_get_persisting_resources_empty_list(self) -> None:
        # Setup
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].return_value = ListStrTemplateOutputDefinition(
            output_name="persisting_resources", value=[]
        )
        mock_manifest = Mock()
        mock_manifest.get_engine.return_value = "terraform"
        handler = EngineDownTester(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Execute
        result = handler.get_persisting_resources()

        # Verify
        self.assertEqual(result, [])
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once_with(
            "persisting_resources", ListStrTemplateOutputDefinition
        )

    def test_get_persisting_resources_wrong_type(self) -> None:
        # Setup
        mock_outputs_handler, mock_outputs_handler_fns = self.get_mock_outputs_handler_and_fns()
        mock_outputs_handler_fns["get_declared_output_def"].side_effect = TypeError("Wrong type")
        mock_manifest = Mock()
        mock_manifest.get_engine.return_value = "terraform"
        handler = EngineDownTester(
            project_path=Path("/project/path"),
            project_manifest=mock_manifest,
            output_handler=mock_outputs_handler,
        )

        # Execute
        result = handler.get_persisting_resources()

        # Verify
        self.assertEqual(result, [])
        mock_outputs_handler_fns["get_declared_output_def"].assert_called_once_with(
            "persisting_resources", ListStrTemplateOutputDefinition
        )



import sys
import unittest
from collections.abc import Generator
from contextlib import contextmanager
from unittest.mock import MagicMock, Mock, patch

from typer.testing import CliRunner

from jupyter_deploy.cli.app import JupyterDeployApp, JupyterDeployCliRunner, main
from jupyter_deploy.cli.app import runner as app_runner
from jupyter_deploy.engine.enum import EngineType


class TestJupyterDeployCliRunner(unittest.TestCase):
    """Test cases for the JupyterDeployCliRunner class."""

    def test_init(self) -> None:
        """Test the initialization of the JupyterDeployCliRunner class."""
        # Create an instance of the class
        runner = JupyterDeployCliRunner()

        self.assertIsNotNone(runner.app, "attribute app should be set")

        # Check that sub-commands are added

        # At least server, host, users, teams, organization
        self.assertGreaterEqual(len(runner.app.registered_groups), 5)
        registered_group_names = [group.name for group in runner.app.registered_groups]
        self.assertIn("server", registered_group_names)
        self.assertIn("host", registered_group_names)
        self.assertIn("users", registered_group_names)
        self.assertIn("teams", registered_group_names)
        self.assertIn("organization", registered_group_names)

    @patch("jupyter_deploy.cli.app.typer.Typer")
    def test_run(self, mock_typer: MagicMock) -> None:
        """Test the run method."""
        # Create a mock app
        mock_app = MagicMock()
        mock_typer.return_value = mock_app

        runner = JupyterDeployCliRunner()
        runner.run()

        # Check that the app was called
        mock_app.assert_called_once()

    def test_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["--help"])

        # Check that the command ran successfully
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(result.stdout.index("Jupyter-deploy") >= 0)
        self.assertTrue(result.stdout.index("server") >= 0)
        self.assertTrue(result.stdout.index("host") >= 0)
        self.assertTrue(result.stdout.index("users") >= 0)
        self.assertTrue(result.stdout.index("teams") >= 0)
        self.assertTrue(result.stdout.index("organization") >= 0)

    def test_no_arg_defaults_to_help(self) -> None:
        runner = CliRunner()
        result = runner.invoke(app_runner.app, [])

        # Check that the command ran successfully
        self.assertEqual(result.exit_code, 0)
        self.assertTrue(result.stdout.index("Jupyter-deploy") >= 0)


class TestJupyterDeployConfigCmd(unittest.TestCase):
    """Test cases for the config method of the JupyterDeployCliRunner class."""

    def get_mock_config_handler(self) -> tuple[Mock, dict[str, Mock]]:
        mock_config_handler = Mock()
        mock_validate = Mock()
        mock_reset_variables = Mock()
        mock_reset_secrets = Mock()
        mock_verify = Mock()
        mock_configure = Mock()
        mock_record = Mock()
        mock_has_used_preset = Mock()

        mock_config_handler.validate_and_set_preset = mock_validate
        mock_config_handler.reset_recorded_variables = mock_reset_variables
        mock_config_handler.reset_recorded_secrets = mock_reset_secrets
        mock_config_handler.verify_requirements = mock_verify
        mock_config_handler.configure = mock_configure
        mock_config_handler.record = mock_record
        mock_config_handler.has_used_preset = mock_has_used_preset

        mock_validate.return_value = True
        mock_verify.return_value = True
        mock_configure.return_value = True
        mock_has_used_preset.return_value = False

        return mock_config_handler, {
            "validate_and_set_preset": mock_validate,
            "reset_recorded_variables": mock_reset_variables,
            "reset_recorded_secrets": mock_reset_secrets,
            "verify": mock_verify,
            "configure": mock_configure,
            "record": mock_record,
            "has_used_preset": mock_has_used_preset,
        }

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_cmd_calls_validate_verify_configure_and_record(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once()
        mock_config_fns["validate_and_set_preset"].assert_called_once_with(
            preset_name="all", will_reset_variables=False
        )
        mock_config_fns["verify"].assert_called_once()
        mock_config_fns["configure"].assert_called_with(variable_overrides={})
        mock_config_fns["record"].assert_called_once_with(record_vars=True, record_secrets=False)
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()
        mock_config_fns["has_used_preset"].assert_called_with("all")

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_passes_all_as_default_preset(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once_with(output_filename=None)
        mock_config_fns["validate_and_set_preset"].assert_called_once_with(
            preset_name="all", will_reset_variables=False
        )
        mock_config_fns["has_used_preset"].assert_called_with("all")

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_passes_no_preset_when_user_passes_none(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "--defaults", "none"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once_with(output_filename=None)
        mock_config_fns["validate_and_set_preset"].assert_called_once_with(preset_name=None, will_reset_variables=False)
        mock_config_fns["has_used_preset"].assert_called_with(None)

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_passes_the_preset_name_when_user_provides_a_value(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "-d", "some-preset"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once_with(output_filename=None)
        mock_config_fns["validate_and_set_preset"].assert_called_once_with(
            preset_name="some-preset", will_reset_variables=False
        )
        mock_config_fns["has_used_preset"].assert_called_with("some-preset")

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_stops_if_validate_returns_false(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance
        mock_config_fns["validate_and_set_preset"].return_value = False

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once()
        mock_config_fns["validate_and_set_preset"].assert_called_once()
        mock_config_fns["verify"].assert_not_called()
        mock_config_fns["configure"].assert_not_called()
        mock_config_fns["record"].assert_not_called()
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()
        mock_config_fns["has_used_preset"].assert_not_called()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_stops_if_verify_requirements_returns_false(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance
        mock_config_fns["verify"].return_value = False

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once()
        mock_config_fns["validate_and_set_preset"].assert_called_once()
        mock_config_fns["verify"].assert_called_once()
        mock_config_fns["configure"].assert_not_called()
        mock_config_fns["record"].assert_not_called()
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()
        mock_config_fns["has_used_preset"].assert_not_called()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_stops_if_configure_returns_false(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance
        mock_config_fns["configure"].return_value = False

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once()
        mock_config_fns["validate_and_set_preset"].assert_called_once()
        mock_config_fns["verify"].assert_called_once()
        mock_config_fns["configure"].assert_called_once()
        mock_config_fns["record"].assert_not_called()
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()
        mock_config_fns["has_used_preset"].assert_not_called()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_reset_vars_and_secrets_when_user_asks(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "--reset"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_fns["validate_and_set_preset"].assert_called_once_with(preset_name="all", will_reset_variables=True)
        mock_config_fns["reset_recorded_variables"].assert_called_once()
        mock_config_fns["reset_recorded_secrets"].assert_called_once()
        mock_config_fns["verify"].assert_called_once()
        mock_config_fns["configure"].assert_called_once()
        mock_config_fns["record"].assert_called_once_with(record_vars=True, record_secrets=False)
        mock_config_fns["has_used_preset"].assert_called_once()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_accepts_r_short_flag_for_reset(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "-r"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_fns["validate_and_set_preset"].assert_called_once_with(preset_name="all", will_reset_variables=True)
        mock_config_fns["record"].assert_called_once_with(record_vars=True, record_secrets=False)
        mock_config_fns["reset_recorded_variables"].assert_called_once()
        mock_config_fns["reset_recorded_secrets"].assert_called_once()
        mock_config_fns["has_used_preset"].assert_called_once()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_with_reset_flag_calls_reset_before_configure_and_record(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        call_order: list[str] = []

        def configure_mock(*a: list, **kw: dict) -> bool:
            call_order.append("configure")
            return True

        mock_config_fns["reset_recorded_variables"].side_effect = lambda *a, **kw: call_order.append("reset_vars")
        mock_config_fns["reset_recorded_secrets"].side_effect = lambda *a, **kw: call_order.append("reset_secrets")
        mock_config_fns["configure"].side_effect = configure_mock
        mock_config_fns["record"].side_effect = lambda *a, **kw: call_order.append("record")

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "-r"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        self.assertEqual(call_order, ["reset_vars", "reset_secrets", "configure", "record"])

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_records_secrets_when_the_user_asks(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "--record-secrets"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_fns["record"].assert_called_once_with(record_vars=True, record_secrets=True)
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_accept_s_flag_to_record_secrets(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "-s"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_fns["record"].assert_called_once_with(record_vars=True, record_secrets=True)
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()

    @patch("jupyter_deploy.handlers.project.config_handler.ConfigHandler")
    def test_config_skip_verify(self, mock_config_handler: Mock) -> None:
        mock_config_handler_instance, mock_config_fns = self.get_mock_config_handler()
        mock_config_handler.return_value = mock_config_handler_instance

        # Act
        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["config", "--skip-verify"])

        # Verify
        self.assertEqual(result.exit_code, 0)
        mock_config_handler.assert_called_once()
        mock_config_fns["validate_and_set_preset"].assert_called_once()
        mock_config_fns["verify"].assert_not_called()
        mock_config_fns["configure"].assert_called_once()
        mock_config_fns["record"].assert_called_once()
        mock_config_fns["reset_recorded_variables"].assert_not_called()
        mock_config_fns["reset_recorded_secrets"].assert_not_called()
        mock_config_fns["has_used_preset"].assert_called_once()


class TestJupyterDeployApp(unittest.TestCase):
    """Test cases for the JupyterDeployApp class."""

    @patch("jupyter_deploy.cli.app.runner")
    def test_start(self, mock_runner: MagicMock) -> None:
        """Test the start method."""
        app = JupyterDeployApp()

        # Test with normal arguments
        with patch.object(sys, "argv", ["jupyter", "deploy", "--help"]):
            app.start()
            mock_runner.run.assert_called_once()
            mock_runner.reset_mock()

        # Test with no arguments
        with patch.object(sys, "argv", ["jupyter", "deploy"]):
            app.start()
            mock_runner.run.assert_called_once()


class TestMain(unittest.TestCase):
    """Test cases for the main function."""

    @patch("jupyter_deploy.cli.app.runner")
    @patch("jupyter_deploy.cli.app.JupyterDeployApp.launch_instance")
    def test_main_as_jupyter_deploy(self, mock_launch_instance: MagicMock, mock_runner: MagicMock) -> None:
        """Test the main function when called as 'jupyter deploy'."""
        with patch.object(sys, "argv", ["jupyter", "deploy"]):
            main()
            mock_launch_instance.assert_called_once()
            mock_runner.run.assert_not_called()

    @patch("jupyter_deploy.cli.app.runner")
    @patch("jupyter_deploy.cli.app.JupyterDeployApp.launch_instance")
    def test_main_as_jupyter_deploy_command(self, mock_launch_instance: MagicMock, mock_runner: MagicMock) -> None:
        """Test the main function when called as 'jupyter-deploy'."""
        with patch.object(sys, "argv", ["jupyter-deploy"]):
            main()
            mock_launch_instance.assert_not_called()
            mock_runner.run.assert_called_once()


class TestInitCommand(unittest.TestCase):
    """Test cases for the init command."""

    def get_mock_project(self) -> Mock:
        """Return a mock project."""
        mock_project = Mock()

        self.mock_may_export_to_project_path = Mock()
        self.mock_clear_project_path = Mock()
        self.mock_setup = Mock()

        self.mock_may_export_to_project_path.return_value = True

        mock_project.may_export_to_project_path = self.mock_may_export_to_project_path
        mock_project.clear_project_path = self.mock_clear_project_path
        mock_project.setup = self.mock_setup

        return mock_project

    @patch("jupyter_deploy.cli.app.InitHandler")
    def test_init_command_no_args_default_to_terraform(self, mock_handler_cls: Mock) -> None:
        """Test that the init command picks up defaults."""
        mock_handler_cls.return_value = self.get_mock_project()

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init", "."])

        # Check that the command ran successfully
        self.assertEqual(result.exit_code, 0, "init command should work")

        mock_handler_cls.assert_called_once_with(
            project_dir=".",
            engine=EngineType.TERRAFORM,
            provider="aws",
            infrastructure="ec2",
            template="base",
        )

    @patch("jupyter_deploy.cli.app.InitHandler")
    def test_init_command_passes_attributes_to_project(self, mock_handler_cls: Mock) -> None:
        """Test that the init command handles optional attributes."""
        mock_handler_cls.return_value = self.get_mock_project()

        runner = CliRunner()
        result = runner.invoke(
            app_runner.app,
            [
                "init",
                "--engine",
                "terraform",
                "--provider",
                "aws",
                "--infrastructure",
                "ec2",
                "--template",
                "other-template",
                "custom-dir",
            ],
        )

        # Check that the command ran successfully
        self.assertEqual(result.exit_code, 0, "init command should work")

        mock_handler_cls.assert_called_once_with(
            project_dir="custom-dir",
            engine=EngineType.TERRAFORM,
            provider="aws",
            infrastructure="ec2",
            template="other-template",
        )

    @patch("jupyter_deploy.cli.app.InitHandler")
    def test_init_command_handles_short_options(self, mock_handler_cls: Mock) -> None:
        """Test that the init command handles short names of optional attributes."""
        mock_handler_cls.return_value = self.get_mock_project()

        runner = CliRunner()
        result = runner.invoke(
            app_runner.app,
            ["init", "-E", "terraform", "-P", "aws", "-I", "ec2", "-T", "a-template", "custom-dir"],
        )

        # Check that the command ran successfully
        self.assertEqual(result.exit_code, 0, "init command should work")

        mock_handler_cls.assert_called_once_with(
            project_dir="custom-dir",
            engine=EngineType.TERRAFORM,
            provider="aws",
            infrastructure="ec2",
            template="a-template",
        )

    @patch("jupyter_deploy.cli.app.InitHandler")
    def test_init_command_calls_project_methods(self, mock_handler_cls: Mock) -> None:
        """Test that the init commands correctly calls project.may_export() and .setup()."""
        mock_handler_cls.return_value = self.get_mock_project()

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init", "."])

        self.assertEqual(result.exit_code, 0, "init command should work")
        self.mock_may_export_to_project_path.assert_called_once()
        self.mock_setup.assert_called_once()

    @patch("jupyter_deploy.cli.app.InitHandler")
    def test_init_command_exits_on_project_conflict_without_overwrite(self, mock_handler_cls: Mock) -> None:
        """Test that the init command exits on existing project conflict when --overwrite is False."""
        mock_handler_cls.return_value = self.get_mock_project()
        self.mock_may_export_to_project_path.return_value = False

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init", "."])

        self.assertEqual(result.exit_code, 0, "init command should work")
        self.mock_may_export_to_project_path.assert_called_once()
        self.mock_clear_project_path.assert_not_called()
        self.mock_setup.assert_not_called()

    @patch("jupyter_deploy.cli.app.InitHandler")
    @patch("jupyter_deploy.cli.app.typer.confirm")
    def test_init_command_with_overwrite_and_user_confirms(self, mock_confirm: Mock, mock_handler_cls: Mock) -> None:
        """Test that the init command with --overwrite prompts the user and proceeds when confirmed."""
        mock_handler_cls.return_value = self.get_mock_project()
        self.mock_may_export_to_project_path.return_value = False
        mock_confirm.return_value = True

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init", "--overwrite", "."])

        self.assertEqual(result.exit_code, 0, "init command should work")
        self.mock_may_export_to_project_path.assert_called_once()
        mock_confirm.assert_called_once()
        self.mock_setup.assert_called_once()

    @patch("jupyter_deploy.cli.app.InitHandler")
    @patch("jupyter_deploy.cli.app.typer.confirm")
    def test_init_command_with_overwrite_and_user_declines(self, mock_confirm: Mock, mock_handler_cls: Mock) -> None:
        """Test that the init command with --overwrite prompts the user and aborts when declined."""
        mock_handler_cls.return_value = self.get_mock_project()
        self.mock_may_export_to_project_path.return_value = False
        mock_confirm.return_value = False

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init", "--overwrite", "."])

        self.assertEqual(result.exit_code, 0, "init command should work")
        self.mock_may_export_to_project_path.assert_called_once()
        mock_confirm.assert_called_once()
        self.mock_setup.assert_not_called()

    @patch("jupyter_deploy.cli.app.InitHandler")
    @patch("jupyter_deploy.cli.app.typer.confirm")
    def test_init_command_with_overwrite_on_no_conflict(self, mock_confirm: Mock, mock_handler_cls: Mock) -> None:
        """Test that the init command with --overwrite proceeds without confirmation if no project conflict."""
        mock_handler_cls.return_value = self.get_mock_project()
        self.mock_may_export_to_project_path.return_value = True

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init", "--overwrite", "."])

        self.assertEqual(result.exit_code, 0, "init command should work")
        self.mock_may_export_to_project_path.assert_called_once()
        mock_confirm.assert_not_called()
        self.mock_setup.assert_called_once()

    @patch("subprocess.run")
    def test_init_command_calls_help_when_no_path(self, mock_subprocess_run: Mock) -> None:
        mock_subprocess_run.return_value = Mock(returncode=0)

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["init"])

        self.assertEqual(result.exit_code, 0, "init command should succeed without path")
        mock_subprocess_run.assert_called_once_with(["jupyter", "deploy", "init", "--help"])


class TestUpCommand(unittest.TestCase):
    def get_mock_up_handler(self) -> tuple[Mock, dict[str, Mock]]:
        mock_up_handler = Mock()
        mock_get_config_file_path = Mock()
        mock_apply = Mock()
        mock_get_default_filename = Mock()

        mock_up_handler.get_config_file_path = mock_get_config_file_path
        mock_up_handler.apply = mock_apply
        mock_up_handler.get_default_config_filename = mock_get_default_filename

        mock_get_default_filename.return_value = "jdout-tfplan"
        mock_get_config_file_path.return_value = ""

        return mock_up_handler, {
            "get_config_file_path": mock_get_config_file_path,
            "apply": mock_apply,
            "get_default_filename": mock_get_default_filename,
        }

    @contextmanager
    def mock_project_dir(*_args: object, **_kwargs: object) -> Generator[None]:
        yield None

    @patch("jupyter_deploy.cli.app.UpHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_up_command_checks_plan_file_exists(
        self, mock_project_ctx_manager: Mock, mock_up_handler_cls: Mock
    ) -> None:
        mock_project_ctx_manager.side_effect = TestUpCommand.mock_project_dir

        mock_up_handler_instance, mock_up_fns = self.get_mock_up_handler()
        mock_up_handler_cls.return_value = mock_up_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["up"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_up_fns["get_config_file_path"].assert_called_once_with(None)

    @patch("jupyter_deploy.cli.app.UpHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_up_command_with_custom_path(self, mock_project_ctx_manager: Mock, mock_up_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestUpCommand.mock_project_dir

        mock_up_handler_instance, mock_up_fns = self.get_mock_up_handler()
        mock_up_handler_cls.return_value = mock_up_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["up", "--path", "/custom/path"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with("/custom/path")

    @patch("jupyter_deploy.cli.app.UpHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_up_command_with_custom_config_file(
        self, mock_project_ctx_manager: Mock, mock_up_handler_cls: Mock
    ) -> None:
        mock_project_ctx_manager.side_effect = TestUpCommand.mock_project_dir

        mock_up_handler_instance, mock_up_fns = self.get_mock_up_handler()
        mock_up_handler_cls.return_value = mock_up_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["up", "--config-filename", "custom-plan"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_up_fns["get_config_file_path"].assert_called_once_with("custom-plan")

    @patch("jupyter_deploy.cli.app.UpHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_up_command_runs_apply_when_config_exists(
        self, mock_project_ctx_manager: Mock, mock_up_handler_cls: Mock
    ) -> None:
        mock_project_ctx_manager.side_effect = TestUpCommand.mock_project_dir

        mock_up_handler_instance, mock_up_fns = self.get_mock_up_handler()
        mock_up_fns["get_config_file_path"].return_value = "/path/to/config"
        mock_up_handler_cls.return_value = mock_up_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["up"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_up_fns["get_config_file_path"].assert_called_once_with(None)
        mock_up_fns["apply"].assert_called_once_with("/path/to/config", False)

    @patch("jupyter_deploy.cli.app.UpHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_up_command_with_answer_yes_option(self, mock_project_ctx_manager: Mock, mock_up_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestUpCommand.mock_project_dir

        mock_up_handler_instance, mock_up_fns = self.get_mock_up_handler()
        mock_up_fns["get_config_file_path"].return_value = "/path/to/config"
        mock_up_handler_cls.return_value = mock_up_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["up", "--answer-yes"])

        self.assertEqual(result.exit_code, 0)
        mock_up_fns["apply"].assert_called_once_with("/path/to/config", True)

    @patch("jupyter_deploy.cli.app.UpHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_up_command_with_all_args(self, mock_project_ctx_manager: Mock, mock_up_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestUpCommand.mock_project_dir

        mock_up_handler_instance, mock_up_fns = self.get_mock_up_handler()
        mock_up_fns["get_config_file_path"].return_value = "/path/to/config"
        mock_up_handler_cls.return_value = mock_up_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["up", "--path", "/custom/path", "--answer-yes"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with("/custom/path")
        mock_up_fns["get_config_file_path"].assert_called_once_with(None)
        mock_up_fns["apply"].assert_called_once_with("/path/to/config", True)


class TestDownCommand(unittest.TestCase):
    def get_mock_down_handler(self) -> tuple[Mock, dict[str, Mock]]:
        mock_down_handler = Mock()
        mock_destroy = Mock()

        mock_down_handler.destroy = mock_destroy

        return mock_down_handler, {"destroy": mock_destroy}

    @contextmanager
    def mock_project_dir(*_args: object, **_kwargs: object) -> Generator[None]:
        yield None

    @patch("jupyter_deploy.cli.app.DownHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_down_command_runs_destroy(self, mock_project_ctx_manager: Mock, mock_down_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestDownCommand.mock_project_dir

        mock_down_handler_instance, mock_down_fns = self.get_mock_down_handler()
        mock_down_handler_cls.return_value = mock_down_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["down"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_down_fns["destroy"].assert_called_once()

    @patch("jupyter_deploy.cli.app.DownHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_down_command_with_custom_path(self, mock_project_ctx_manager: Mock, mock_down_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestDownCommand.mock_project_dir

        mock_down_handler_instance, mock_down_fns = self.get_mock_down_handler()
        mock_down_handler_cls.return_value = mock_down_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["down", "--path", "/custom/path"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with("/custom/path")
        mock_down_fns["destroy"].assert_called_once()

    @patch("jupyter_deploy.cli.app.DownHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_down_command_with_answer_yes_option(
        self, mock_project_ctx_manager: Mock, mock_down_handler_cls: Mock
    ) -> None:
        mock_project_ctx_manager.side_effect = TestDownCommand.mock_project_dir

        mock_down_handler_instance, mock_down_fns = self.get_mock_down_handler()
        mock_down_handler_cls.return_value = mock_down_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["down", "--answer-yes"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_down_fns["destroy"].assert_called_once_with(True)


class TestOpenCommand(unittest.TestCase):
    @contextmanager
    def mock_project_dir(*_args: object, **_kwargs: object) -> Generator[None]:
        yield None

    def get_mock_open_handler(self) -> tuple[Mock, dict[str, Mock]]:
        mock_open_handler = Mock()
        mock_open_url = Mock()

        mock_open_handler.open = mock_open_url
        return mock_open_handler, {"open": mock_open_url}

    @patch("jupyter_deploy.cli.app.OpenHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_open_command_runs_open(self, mock_project_ctx_manager: Mock, mock_open_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestOpenCommand.mock_project_dir

        mock_open_handler_instance, mock_open_fns = self.get_mock_open_handler()
        mock_open_handler_cls.return_value = mock_open_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["open"])

        assert result.exit_code == 0
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_open_fns["open"].assert_called_once()

    @patch("jupyter_deploy.cli.app.OpenHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_open_command_with_custom_path(self, mock_project_ctx_manager: Mock, mock_open_handler_cls: Mock) -> None:
        mock_project_ctx_manager.side_effect = TestOpenCommand.mock_project_dir

        mock_open_handler_instance, mock_open_fns = self.get_mock_open_handler()
        mock_open_handler_cls.return_value = mock_open_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["open", "--path", "/custom/path"])

        assert result.exit_code == 0
        mock_project_ctx_manager.assert_called_once_with("/custom/path")
        mock_open_fns["open"].assert_called_once()


class TestShowCommand(unittest.TestCase):
    """Test cases for the show command."""

    @contextmanager
    def mock_project_dir(*_args: object, **_kwargs: object) -> Generator[None]:
        yield None

    def get_mock_show_handler(self) -> tuple[Mock, dict[str, Mock]]:
        mock_show_handler = Mock()
        mock_show_project_info = Mock()

        mock_show_handler.show_project_info = mock_show_project_info

        return mock_show_handler, {"show_project_info": mock_show_project_info}

    @patch("jupyter_deploy.cli.app.ShowHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_show_command_default_flags(self, mock_project_ctx_manager: Mock, mock_show_handler_cls: Mock) -> None:
        """Test that show command with no flags shows all sections."""
        mock_project_ctx_manager.side_effect = TestShowCommand.mock_project_dir

        mock_show_handler_instance, mock_show_fns = self.get_mock_show_handler()
        mock_show_handler_cls.return_value = mock_show_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["show"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_show_fns["show_project_info"].assert_called_once_with(
            show_info=True, show_outputs=True, show_variables=True
        )

    @patch("jupyter_deploy.cli.app.ShowHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_show_command_with_info_flag(self, mock_project_ctx_manager: Mock, mock_show_handler_cls: Mock) -> None:
        """Test that show command with --info flag shows only info section."""
        mock_project_ctx_manager.side_effect = TestShowCommand.mock_project_dir

        mock_show_handler_instance, mock_show_fns = self.get_mock_show_handler()
        mock_show_handler_cls.return_value = mock_show_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["show", "--info"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_show_fns["show_project_info"].assert_called_once_with(
            show_info=True, show_outputs=False, show_variables=False
        )

    @patch("jupyter_deploy.cli.app.ShowHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_show_command_with_outputs_flag(self, mock_project_ctx_manager: Mock, mock_show_handler_cls: Mock) -> None:
        """Test that show command with --outputs flag shows only outputs section."""
        mock_project_ctx_manager.side_effect = TestShowCommand.mock_project_dir

        mock_show_handler_instance, mock_show_fns = self.get_mock_show_handler()
        mock_show_handler_cls.return_value = mock_show_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["show", "--outputs"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_show_fns["show_project_info"].assert_called_once_with(
            show_info=False, show_outputs=True, show_variables=False
        )

    @patch("jupyter_deploy.cli.app.ShowHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_show_command_with_variables_flag(
        self, mock_project_ctx_manager: Mock, mock_show_handler_cls: Mock
    ) -> None:
        """Test that show command with --variables flag shows only variables section."""
        mock_project_ctx_manager.side_effect = TestShowCommand.mock_project_dir

        mock_show_handler_instance, mock_show_fns = self.get_mock_show_handler()
        mock_show_handler_cls.return_value = mock_show_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["show", "--variables"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_show_fns["show_project_info"].assert_called_once_with(
            show_info=False, show_outputs=False, show_variables=True
        )

    @patch("jupyter_deploy.cli.app.ShowHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_show_command_with_multiple_flags(
        self, mock_project_ctx_manager: Mock, mock_show_handler_cls: Mock
    ) -> None:
        """Test that show command with multiple flags shows only selected sections."""
        mock_project_ctx_manager.side_effect = TestShowCommand.mock_project_dir

        mock_show_handler_instance, mock_show_fns = self.get_mock_show_handler()
        mock_show_handler_cls.return_value = mock_show_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["show", "--info", "--outputs"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with(None)
        mock_show_fns["show_project_info"].assert_called_once_with(
            show_info=True, show_outputs=True, show_variables=False
        )

    @patch("jupyter_deploy.cli.app.ShowHandler")
    @patch("jupyter_deploy.cmd_utils.project_dir")
    def test_show_command_with_custom_path(self, mock_project_ctx_manager: Mock, mock_show_handler_cls: Mock) -> None:
        """Test show command with custom project path."""
        mock_project_ctx_manager.side_effect = TestShowCommand.mock_project_dir

        mock_show_handler_instance, mock_show_fns = self.get_mock_show_handler()
        mock_show_handler_cls.return_value = mock_show_handler_instance

        runner = CliRunner()
        result = runner.invoke(app_runner.app, ["show", "--path", "/custom/path"])

        self.assertEqual(result.exit_code, 0)
        mock_project_ctx_manager.assert_called_once_with("/custom/path")
        mock_show_fns["show_project_info"].assert_called_once_with(
            show_info=True, show_outputs=True, show_variables=True
        )



from pathlib import Path
from typing import Any

from jupyter_deploy import fs_utils
from jupyter_deploy.engine.engine_variables import EngineVariablesHandler
from jupyter_deploy.engine.terraform import tf_varfiles
from jupyter_deploy.engine.terraform.tf_constants import (
    TF_CUSTOM_PRESET_FILENAME,
    TF_ENGINE_DIR,
    TF_PRESETS_DIR,
    TF_RECORDED_SECRETS_FILENAME,
    TF_RECORDED_VARS_FILENAME,
    TF_VARIABLES_FILENAME,
    get_preset_filename,
)
from jupyter_deploy.engine.vardefs import TemplateVariableDefinition
from jupyter_deploy.manifest import JupyterDeployManifest


class TerraformVariablesHandler(EngineVariablesHandler):
    """Terraform-specific implementation of the VariableHandler."""

    def __init__(self, project_path: Path, project_manifest: JupyterDeployManifest) -> None:
        super().__init__(project_path=project_path, project_manifest=project_manifest)
        self._template_vars: dict[str, TemplateVariableDefinition] | None = None
        self.engine_dir_path = project_path / TF_ENGINE_DIR

    def get_recorded_variables_filepath(self) -> Path:
        return self.engine_dir_path / TF_RECORDED_VARS_FILENAME

    def get_recorded_secrets_filepath(self) -> Path:
        return self.engine_dir_path / TF_RECORDED_SECRETS_FILENAME

    def is_template_directory(self) -> bool:
        return fs_utils.file_exists(self.engine_dir_path / TF_VARIABLES_FILENAME)

    def get_template_variables(self) -> dict[str, TemplateVariableDefinition]:
        # cache handling to avoid the expensive fs operation necessary
        # to retrieve the variable definitions.
        if self._template_vars:
            return self._template_vars

        # read the variables.tf, retrieve the description, sensitive
        variables_dot_tf_path = self.engine_dir_path / TF_VARIABLES_FILENAME
        variables_dot_tf_content = fs_utils.read_short_file(variables_dot_tf_path)
        variable_defs = tf_varfiles.parse_variables_dot_tf_content(variables_dot_tf_content)

        # read the template .tfvars with the defaults
        all_defaults_tfvars_path = self.engine_dir_path / TF_PRESETS_DIR / get_preset_filename()
        variables_tfvars_content = fs_utils.read_short_file(all_defaults_tfvars_path)

        # combine
        tf_varfiles.parse_dot_tfvars_content_and_add_defaults(variables_tfvars_content, variable_defs=variable_defs)

        # translate to the engine-generic type
        template_vars = {var_name: var_def.to_template_definition() for var_name, var_def in variable_defs.items()}
        self._template_vars = template_vars
        return template_vars

    def update_variable_records(self, varvalues: dict[str, Any], sensitive: bool = False) -> None:
        if not varvalues:
            return

        template_vars = self.get_template_variables()

        # first verify
        updated_vals: dict[str, Any] = {}
        for varname, varvalue in varvalues.items():
            existing_vardef = template_vars.get(varname)

            if not existing_vardef:
                raise KeyError(f"Variable not found: {varname}")
            converted_value = existing_vardef.validate_value(varvalue)

            # here we leverage pydantic to cast the value.
            # say a variable is an int, and for some reason a command result
            # returned a string "30", pydantic will convert it to 30 automatically.
            updated_vals[varname] = converted_value

        # if all pass, assign
        for varname in varvalues:
            existing_vardef = template_vars[varname]
            existing_vardef.assigned_value = updated_vals[varname]

        # update the .tfvars file, or create a new one if it doesn't exist.
        file_name = TF_RECORDED_VARS_FILENAME if not sensitive else TF_RECORDED_SECRETS_FILENAME
        tfvars_path = self.engine_dir_path / file_name
        previous_tfvars_content: str = ""
        if fs_utils.file_exists(tfvars_path):
            previous_tfvars_content = fs_utils.read_short_file(tfvars_path)

        updated_tfvars_lines = tf_varfiles.parse_and_update_dot_tfvars_content(previous_tfvars_content, varvalues)

        if updated_tfvars_lines:
            fs_utils.write_inline_file_content(tfvars_path, updated_tfvars_lines)

    def create_filtered_preset_file(self, base_preset_path: Path) -> Path:
        """Read the base preset, override values, write in a new preset file and return its path."""
        filtered_tfvars_file_path = self.project_path / TF_ENGINE_DIR / TF_CUSTOM_PRESET_FILENAME

        base_preset_content = fs_utils.read_short_file(base_preset_path)
        assigned_variable_names = self.get_variable_names_assigned_in_config()
        updated_tfvars_lines = tf_varfiles.parse_and_remove_overridden_variables_from_content(
            base_preset_content, assigned_variable_names
        )

        if updated_tfvars_lines:
            fs_utils.write_inline_file_content(filtered_tfvars_file_path, updated_tfvars_lines)

        return filtered_tfvars_file_path

    def reset_recorded_variables(self) -> None:
        super().reset_recorded_variables()

        path = self.get_recorded_variables_filepath()
        deleted = fs_utils.delete_file_if_exists(path)

        if deleted:
            console = self.get_console()
            console.print(f":wastebasket: Deleted previously recorded inputs at: {path.name}")

    def reset_recorded_secrets(self) -> None:
        super().reset_recorded_secrets()

        path = self.get_recorded_secrets_filepath()
        deleted = fs_utils.delete_file_if_exists(path)

        if deleted:
            console = self.get_console()
            console.print(f":wastebasket: Deleted previously recorded secrets at: {path.name}")



import unittest
from unittest.mock import Mock, call, patch

import botocore.exceptions
from mypy_boto3_ssm.literals import CommandInvocationStatusType

from jupyter_deploy.api.aws.ssm.ssm_command import (
    is_terminal_command_invocation_status,
    poll_command,
    send_cmd_to_one_instance_and_wait_sync,
)


class TestIsTerminalCommandInvocationStatus(unittest.TestCase):
    def test_terminal_statuses(self) -> None:
        terminal_statuses: list[CommandInvocationStatusType] = ["Cancelled", "Failed", "Success", "TimedOut"]
        for status in terminal_statuses:
            with self.subTest(status=status):
                self.assertTrue(is_terminal_command_invocation_status(status))

    def test_non_terminal_statuses(self) -> None:
        non_terminal_statuses: list[CommandInvocationStatusType] = ["Cancelling", "Delayed", "InProgress", "Pending"]
        for status in non_terminal_statuses:
            with self.subTest(status=status):
                self.assertFalse(is_terminal_command_invocation_status(status))


class TestPollCommand(unittest.TestCase):
    @patch("time.sleep")
    def test_sleeps_and_retries_when_cmd_not_registered_yet(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.get_command_invocation.side_effect = [
            botocore.exceptions.ClientError(
                {"Error": {"Code": "InvocationDoesNotExist", "Message": "Command not found"}},
                "GetCommandInvocation",
            ),
            {"Status": "Success"},
        ]

        # Execute
        result = poll_command(mock_client, "cmd-123", "i-123", wait_on_invocation_does_not_exist=5)

        # Assert
        self.assertEqual(result["Status"], "Success")
        mock_sleep.assert_called_once_with(5)
        mock_client.get_command_invocation.assert_has_calls(
            [
                call(CommandId="cmd-123", InstanceId="i-123"),
                call(CommandId="cmd-123", InstanceId="i-123"),
            ]
        )

    @patch("time.sleep")
    def test_retries_only_once_when_cmd_not_registered(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.get_command_invocation.side_effect = [
            botocore.exceptions.ClientError(
                {"Error": {"Code": "InvocationDoesNotExist", "Message": "Command not found"}},
                "GetCommandInvocation",
            ),
            botocore.exceptions.ClientError(
                {"Error": {"Code": "InvocationDoesNotExist", "Message": "Command not found"}},
                "GetCommandInvocation",
            ),
        ]

        # Execute & Assert
        with self.assertRaises(botocore.exceptions.ClientError):
            poll_command(mock_client, "cmd-123", "i-123")

        mock_sleep.assert_called_once()

    @patch("time.sleep")
    def test_poll_until_succeeded(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.get_command_invocation.side_effect = [
            {"Status": "Pending"},
            {"Status": "InProgress"},
            {"Status": "Success"},
        ]

        # Execute
        result = poll_command(mock_client, "cmd-123", "i-123", poll_interval_seconds=1)

        # Assert
        self.assertEqual(result["Status"], "Success")
        mock_sleep.assert_has_calls([call(1), call(1)])
        self.assertEqual(mock_client.get_command_invocation.call_count, 3)

    @patch("time.sleep")
    def test_stop_polling_on_failed(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.get_command_invocation.side_effect = [
            {"Status": "InProgress"},
            {"Status": "Failed"},
        ]

        # Execute
        result = poll_command(mock_client, "cmd-123", "i-123")

        # Assert
        self.assertEqual(result["Status"], "Failed")
        mock_sleep.assert_called_once()
        self.assertEqual(mock_client.get_command_invocation.call_count, 2)

    @patch("time.sleep")
    def test_raises_when_get_invocation_status_fails_in_loop(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.get_command_invocation.side_effect = [
            {"Status": "InProgress"},
            botocore.exceptions.ClientError(
                {"Error": {"Code": "SomeError", "Message": "Some error occurred"}},
                "GetCommandInvocation",
            ),
        ]

        # Execute & Assert
        with self.assertRaises(botocore.exceptions.ClientError):
            poll_command(mock_client, "cmd-123", "i-123")

        mock_sleep.assert_called_once()
        self.assertEqual(mock_client.get_command_invocation.call_count, 2)


class TestSendCmdToOneInstanceAndWaitSync(unittest.TestCase):
    @patch("jupyter_deploy.api.aws.ssm.ssm_command.poll_command")
    @patch("time.sleep")
    def test_sends_command(self, mock_sleep: Mock, mock_poll_command: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.return_value = {"Command": {"CommandId": "cmd-123"}}
        mock_poll_command.return_value = {"Status": "Success"}

        # Execute
        result = send_cmd_to_one_instance_and_wait_sync(mock_client, "AWS-RunShellScript", "i-123")

        # Assert
        mock_client.send_command.assert_called_once_with(
            DocumentName="AWS-RunShellScript",
            InstanceIds=["i-123"],
            TimeoutSeconds=30,
        )
        self.assertEqual(result["Status"], "Success")

    @patch("jupyter_deploy.api.aws.ssm.ssm_command.poll_command")
    @patch("time.sleep")
    def test_passes_parameters(self, mock_sleep: Mock, mock_poll_command: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.return_value = {"Command": {"CommandId": "cmd-123"}}
        mock_poll_command.return_value = {"Status": "Success"}

        # Execute
        result = send_cmd_to_one_instance_and_wait_sync(
            mock_client,
            "AWS-RunShellScript",
            "i-123",
            timeout_seconds=20,
            wait_after_send_seconds=1,
            ParamKey1=["users"],
            ParamKey2=["username1", "username2"],
        )

        # Assert
        mock_client.send_command.assert_called_once_with(
            DocumentName="AWS-RunShellScript",
            InstanceIds=["i-123"],
            TimeoutSeconds=20,
            Parameters={
                "ParamKey1": ["users"],
                "ParamKey2": ["username1", "username2"],
            },
        )
        mock_sleep.assert_called_once_with(1)
        self.assertEqual(result["Status"], "Success")

    @patch("jupyter_deploy.api.aws.ssm.ssm_command.poll_command")
    @patch("time.sleep")
    def test_sleeps_before_polling(self, mock_sleep: Mock, mock_poll_command: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.return_value = {"Command": {"CommandId": "cmd-123"}}
        mock_poll_command.return_value = {"Status": "Success"}

        # Execute
        send_cmd_to_one_instance_and_wait_sync(mock_client, "AWS-RunShellScript", "i-123", wait_after_send_seconds=5)

        # Assert
        mock_sleep.assert_called_once_with(5)

    @patch("jupyter_deploy.api.aws.ssm.ssm_command.poll_command")
    @patch("time.sleep")
    def test_polls_and_surface_status(self, mock_sleep: Mock, mock_poll_command: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.return_value = {"Command": {"CommandId": "cmd-123"}}
        mock_poll_command.return_value = {"Status": "Failed", "StandardErrorContent": "Error"}

        # Execute
        result = send_cmd_to_one_instance_and_wait_sync(mock_client, "AWS-RunShellScript", "i-123")

        # Assert
        mock_poll_command.assert_called_once_with(mock_client, command_id="cmd-123", instance_id="i-123")
        self.assertEqual(result["Status"], "Failed")
        self.assertEqual(result["StandardErrorContent"], "Error")

    @patch("time.sleep")
    def test_raises_if_send_command_raises(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.side_effect = botocore.exceptions.ClientError(
            {"Error": {"Code": "SomeError", "Message": "Some error occurred"}},
            "SendCommand",
        )

        # Execute & Assert
        with self.assertRaises(botocore.exceptions.ClientError):
            send_cmd_to_one_instance_and_wait_sync(mock_client, "AWS-RunShellScript", "i-123")

    @patch("time.sleep")
    def test_raises_if_poll_command_raises(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.return_value = {"Command": {"CommandId": "cmd-123"}}

        with patch("jupyter_deploy.api.aws.ssm.ssm_command.poll_command") as mock_poll_command:
            mock_poll_command.side_effect = botocore.exceptions.ClientError(
                {"Error": {"Code": "SomeError", "Message": "Some error occurred"}},
                "GetCommandInvocation",
            )

            # Execute & Assert
            with self.assertRaises(botocore.exceptions.ClientError):
                send_cmd_to_one_instance_and_wait_sync(mock_client, "AWS-RunShellScript", "i-123")

    @patch("time.sleep")
    def test_raises_if_command_id_missing(self, mock_sleep: Mock) -> None:
        # Setup
        mock_client = Mock()
        mock_client.send_command.return_value = {"Command": {}}  # No CommandId

        # Execute & Assert
        with self.assertRaises(RuntimeError) as context:
            send_cmd_to_one_instance_and_wait_sync(mock_client, "AWS-RunShellScript", "i-123")

        self.assertEqual(str(context.exception), "Command ID could not be retrieved.")
